"""Merge databases."""

import pathlib
import sqlite3

import click
from context_verbose import Printer

from mendevi.database.create import create_database

from .parse import parse_videos_database


def _insert_update(
    cursor: sqlite3.Cursor, row: sqlite3.Row | dict, tab: str, keys: tuple[str],
) -> dict[str]:
    """Set or update the tab of the cursor with the new line."""
    line: dict[str] = {k: v for k, v in dict(row).items() if v is not None}
    ids = tuple(line.pop(k) for k in keys)
    line_keys = tuple(line)  # frozen the order
    line_values = tuple(line[k] for k in line_keys)
    try:
        out = cursor.execute(
            (
                f"INSERT INTO {tab} ({', '.join(keys + line_keys)}) "
                f"VALUES ({', '.join('?'*len(keys + line_keys))}) "
                "RETURNING *"
            ),
            ids + line_values,
        ).fetchone()
    except sqlite3.IntegrityError:  # if exists
        cursor.execute(
            (
                f"UPDATE {tab} SET {', '.join(f'{k}=?' for k in line_keys)} "
                f"WHERE {' AND '.join(f'{k}=?' for k in keys)}"
            ),
            line_values + ids,
        )
        out = cursor.execute(
            f"SELECT * FROM {tab} WHERE {' AND '.join(f'{k}=?' for k in keys)}",
            ids,
        ).fetchone()
    return dict(out)


@click.command()
@click.argument("databases", type=click.Path(), nargs=-1)
def main(databases: tuple[str]) -> None:
    """Merge several SQL databases together.

    \b
    You can test if the database is corrupted with:
    ``sqlite3 mendevi.db "PRAGMA integrity_check;"``.
    If is raises "ok", it means the database is corrupted.
    You can fix it with ``sqlite3 mendevi.db ".recover" | sqlite3 mendevi_recovered.db``.

    \b
    Parameters
    ----------
    databases : list[pathlike]
        All databases to be merged.

    """
    with Printer("Parse configuration...") as prt:
        assert isinstance(databases, (tuple, list, set, frozenset)), databases.__class__.__name__
        all_databases: list[pathlib.Path] = []
        for i, database_ in enumerate(databases):
            _, database = parse_videos_database(prt, (), database_, _quiet=True)
            all_databases.append(database)
            prt.print(f"database {i+1:>2}: {database}")
        dst = pathlib.Path(f"merge_{'_'.join(d.stem for d in all_databases)}.db")
        create_database(dst)
        prt.print(f"destination: {dst} (just created)")

    with sqlite3.connect(dst) as conn_dst:
        conn_dst.row_factory = sqlite3.Row
        cursor_dst = conn_dst.cursor()
        for database in all_databases:
            with (
                Printer(f"Add {database.name} into {dst.name}...", color="cyan") as prt,
                sqlite3.connect(f"file:{database}?mode=ro", uri=True) as conn_src,
            ):
                conn_src.row_factory = sqlite3.Row
                cursor_src = conn_src.cursor()

                # t_vid_video
                prt.print("table 't_vid_video'")
                for row_ in cursor_src.execute("SELECT * FROM t_vid_video"):
                    _insert_update(cursor_dst, row_, "t_vid_video", ("vid_id",))

                # t_met_metric
                prt.print("table 't_met_metric'")
                for row_ in cursor_src.execute("SELECT * FROM t_met_metric"):
                    row = dict(row_)
                    del row["met_id"]
                    _insert_update(
                        cursor_dst, row, "t_met_metric", ("met_ref_vid_id", "met_dis_vid_id"),
                    )

                # t_act_activity
                prt.print("table 't_act_activity'")
                act_old_to_new: dict[int, int] = {}
                for row_ in cursor_src.execute("SELECT * FROM t_act_activity"):
                    row = dict(row_)
                    old = row.pop("act_id")
                    new = _insert_update(cursor_dst, row, "t_act_activity", ())["act_id"]
                    act_old_to_new[old] = new

                # t_env_environment
                prt.print("table 't_env_environment'")
                env_old_to_new: dict[int, int] = {}
                for row_ in cursor_src.execute("SELECT * FROM t_env_environment"):
                    row = dict(row_)
                    row["env_idle_act_id"] = act_old_to_new[row["env_idle_act_id"]]
                    old = row.pop("env_id")
                    new = _insert_update(cursor_dst, row, "t_env_environment", ())["env_id"]
                    env_old_to_new[old] = new

                # t_enc_encode
                prt.print("table 't_enc_encode'")
                for row_ in cursor_src.execute("SELECT * FROM t_enc_encode"):
                    row = dict(row_)
                    del row["enc_id"]
                    if "enc_vbr" in row:  # retrocompatibility with version <= 1.2.0
                        row["enc_mode"] = "vbr" if row.pop("enc_vbr") else "cbr"
                    row["enc_env_id"] = env_old_to_new[row["enc_env_id"]]
                    row["enc_act_id"] = act_old_to_new[row["enc_act_id"]]
                    _insert_update(cursor_dst, row, "t_enc_encode", ())

                # t_dec_decode
                prt.print("table 't_dec_decode'")
                for row_ in cursor_src.execute("SELECT * FROM t_dec_decode"):
                    row = dict(row_)
                    del row["dec_id"]
                    row["dec_env_id"] = env_old_to_new[row["dec_env_id"]]
                    row["dec_act_id"] = act_old_to_new[row["dec_act_id"]]
                    _insert_update(cursor_dst, row, "t_dec_decode", ())



"""The encoding and decoding ffmpeg cmd."""

import functools
import math
import re
import subprocess

from mendevi.utils import best_profile


def decorator_pix_fmt(func: callable) -> callable:
    """Add a 8 bit pixel format conversion if required."""
    @functools.wraps(func)
    def decorated_func(**kwargs: dict) -> list:
        cmd = func(**kwargs)
        if (
            kwargs["pix_fmt"] != "yuv420p"
            and not support_pix_fmt(kwargs["encoder"], kwargs["pix_fmt"])
        ):
            cmd = ["-pix_fmt", "yuv420p", *cmd]
        return cmd
    return decorated_func


def quality_to_rate(kwargs: dict[str]) -> int:
    """Return the absolute target bitrate in kbit/s.

    Based on https://twitch-overlay.fr/quelle-connexion-internet-choisir-pour-streamer-sur-twitch/
    and https://bitmovin.com/blog/video-bitrate-streaming-hls-dash/

    You can plot the bitrate with: mendevi plot mendevi.db -x bitrate -y psnr -f 'mode = "vbr"'

    The flow margin is taken to be twice as small and twice as large as the recommendations.
    """
    quality = kwargs["quality"]
    assert isinstance(quality, float), quality.__class__.__name__
    assert 0.0 <= quality <= 1.0, quality
    match (profile := best_profile(*kwargs["resolution"])):
        case "sd":
            mini, maxi = 400, 2100
        case "hd":
            mini, maxi = 1500, 6000
        case "fhd":
            mini, maxi = 3000, 9000
        case "uhd4k":
            mini, maxi = 10000, 51000
        case _:
            msg = f"please define a bitrate rule for the profile {profile}"
            raise NotImplementedError(msg)
    mini, maxi = mini // 2, maxi * 2  # apply margin
    mini, maxi = math.log10(float(mini)), math.log10(float(maxi))
    return round(10.0**(maxi-quality*(maxi-mini)))


@functools.cache
def support_pix_fmt(encoder: str, pix_fmt: str) -> bool:
    """Return True if the encoder supports the given pixel format."""
    cmd = ["ffmpeg", "-hide_banner", "-h", f"encoder={encoder}"]
    res: str = subprocess.run(cmd, check=True, capture_output=True).stdout.decode()
    return re.search(fr"\s{pix_fmt}\s", res) is not None


@decorator_pix_fmt
def _encode_av1_nvenc(**kwargs: dict) -> list[str]:
    """Return the ffmpeg arguments."""
    general = [
        "-c:v", "av1_nvenc",
        "-gpu", "any",
        "-tune", "hq",
        "-preset", {"fast": "p2", "medium": "p4", "slow": "p6"}[kwargs["effort"]],
    ]
    if kwargs["mode"] == "vbr":
        return [
            *general,
            "-rc", "vbr",
            "-cq", str(round(1.0 + kwargs["quality"]*60.0)),  # [1, 63]
        ]
    rate = f"{quality_to_rate(kwargs)}k"
    return [
        *general,
        "-b:v", rate,
        "-minrate", rate,
        "-maxrate", rate,
        "-bufsize", rate,
        "-rc", "cbr",
    ]


def _encode_av1_vaapi(**kwargs: dict) -> list[str]:
    """Return the ffmpeg arguments."""
    return _encode_vaapi("av1", **kwargs)


@decorator_pix_fmt
def _encode_h264_nvenc(**kwargs: dict) -> list[str]:
    """Return the ffmpeg arguments."""
    general = [
        "-c:v", "h264_nvenc",
        "-gpu", "any",
        "-tune", "hq",
        "-preset", {"fast": "p2", "medium": "p4", "slow": "p6"}[kwargs["effort"]],
    ]
    if kwargs["mode"] == "vbr":
        return [
            *general,
            "-rc", "vbr",
            "-cq", str(round(1.0 + kwargs["quality"]*50.0)),  # [1, 51]
        ]
    rate = f"{quality_to_rate(kwargs)}k"
    return [
        *general,
        "-b:v", rate,
        "-minrate", rate,
        "-maxrate", rate,
        "-bufsize", rate,
        "-rc", "cbr",
    ]


def _encode_vaapi(codec: str, **kwargs: dict) -> list[str]:
    """Return the ffmpeg arguments for the xxx_vaapi encoders."""
    # https://trac.ffmpeg.org/wiki/Hardware/VAAPI
    # https://www.ffmpeg.org/ffmpeg-codecs.html#VAAPI-encoders
    # see compatibility with "vainfo --all"
    pix_fmt = {
        "yuv420p": "nv12",
        "yuv420p10le": "p010",
    }[kwargs["pix_fmt"]]
    compression = {
        "slow": "1",
        "medium": "3",
        "fast": "6",
    }[kwargs["effort"]]
    general = [
        "-vaapi_device", "/dev/dri/renderD128",
        "-vf", f"format={pix_fmt},hwupload",
        "-c:v", f"{codec}_vaapi",
        "-async_depth", str(kwargs["threads"]),
        "-compression_level", compression,
    ]
    if kwargs["mode"] == "vbr":
        return [
            *general,
            # "rc_mode", "CQP",  # very rare VBR support
            "-qp", str(round(1.0 + kwargs["quality"]*51.0)),  # [1, 52]
        ]
    rate = f"{quality_to_rate(kwargs)}k"
    return [
        *general,
        "-b:v", rate,
        "-minrate", rate,
        "-maxrate", rate,
        "-bufsize", rate,
        # "rc_mode", "CBR",
    ]


def _encode_h264_vaapi(**kwargs: dict) -> list[str]:
    """Return the ffmpeg arguments."""
    return _encode_vaapi("h264", **kwargs)


@decorator_pix_fmt
def _encode_hevc_nvenc(**kwargs: dict) -> list[str]:
    """Return the ffmpeg arguments."""
    general = [
        "-c:v", "hevc_nvenc",
        "-gpu", "any",
        "-tune", "hq",
        "-preset", {"fast": "p2", "medium": "p4", "slow": "p6"}[kwargs["effort"]],
    ]
    if kwargs["mode"] == "vbr":
        return [
            *general,
            "-rc", "vbr",
            "-cq", str(round(1.0 + kwargs["quality"]*50.0)),  # [1, 51]
        ]
    rate = f"{quality_to_rate(kwargs)}k"
    return [
        *general,
        "-b:v", rate,
        "-minrate", rate,
        "-maxrate", rate,
        "-bufsize", rate,
        "-rc", "cbr",
    ]


def _encode_hevc_vaapi(**kwargs: dict) -> list[str]:
    """Return the ffmpeg arguments."""
    return _encode_vaapi("hevc", **kwargs)


@decorator_pix_fmt
def _encode_libaomav1(**kwargs: dict) -> list[str]:
    """Return the ffmpeg arguments."""
    # find smart tiles blocs
    columns = math.ceil(math.sqrt(kwargs["threads"]))
    rows = math.ceil(kwargs["threads"]/columns)
    tiles = f"{columns}x{rows}"
    general = [
        "-c:v", "libaom-av1",
        "-cpu-used", {"fast": "7", "medium": "4", "slow": "1"}[kwargs["effort"]],
        "-tune", "ssim",
        "-threads", str(kwargs["threads"]), "-row-mt", "1", "-tiles", tiles,
        "-denoise-noise-level", "0",
    ]
    if kwargs["mode"] == "vbr":
        return [
            *general,
            "-crf", str(round(kwargs["quality"]*63.0)),  # in [0, 63]
        ]
    rate = f"{quality_to_rate(kwargs)}k"
    return [
        *general,
        "-b:v", rate,
        "-minrate", rate,
        "-maxrate", rate,
        "-bufsize", rate,
    ]


@decorator_pix_fmt
def _encode_libopenh264(**kwargs: dict) -> list[str]:
    """Return the ffmpeg arguments."""
    general = [
        "-c:v", "libopenh264",
        "-profile", {
            "fast": "constrained_baseline", "medium": "main", "slow": "high",
        }[kwargs["effort"]],
        "-threads", str(kwargs["threads"]), "-slices", str(kwargs["threads"]),
        # "-slice_mode", "dyn",
    ]
    if kwargs["effort"] != "medium":
        general = [
            *general,
            "-loopfilter", {"slow": "1", "fast": "0"}[kwargs["effort"]],
        ]
    if kwargs["mode"] == "vbr":
        # https://patchwork.ffmpeg.org/project/ffmpeg/patch/
        # 1585926759-22569-1-git-send-email-linjie.fu@intel.com/
        quantization = str(round(1.0 + kwargs["quality"]*50.0))  # [1, 51]
        return [
            *general,
            "-qmin", quantization,
            "-qmax", quantization,
            "-rc_mode", "quality",
        ]
    rate = f"{quality_to_rate(kwargs)}k"
    return [
        *general,
        "-b:v", rate,
        "-minrate", rate,
        "-maxrate", rate,
        "-bufsize", rate,
        "-rc_mode", "bitrate",
    ]


@decorator_pix_fmt
def _encode_librav1e(**kwargs: dict) -> list[str]:
    """Return the ffmpeg arguments."""
    # https://docs.rs/rav1e/latest/rav1e/config/struct.EncoderConfig.html
    if kwargs["mode"] == "vbr":
        quality = ["-qp", str(round(kwargs["quality"]*255))]  # not realy constant quality
    else:
        rate = f"{quality_to_rate(kwargs)}k"
        quality = ["-b:v", rate, "-minrate", rate, "-maxrate", rate]
    return [
        "-c:v", "librav1e",
        *quality,
        # speed in 0, 10, default 6
        "-speed", {"slow": "2", "medium": "6", "fast": "9"}[kwargs["effort"]],
        "-rav1e-params", (
            f"threads={kwargs['threads']}:tiles={kwargs['threads']}:photon-noise=0"
        ),
    ]


@decorator_pix_fmt
def _encode_libsvtav1(**kwargs: dict) -> list[str]:
    """Return the ffmpeg arguments."""
    def libsvtav1_lp(threads: int) -> int:
        """Convert threads in parralel level."""
        # https://gitlab.com/AOMediaCodec/SVT-AV1/-/blob/master/Source/Lib/Globals/enc_handle.c#L613
        # lp=1 -> threads=1
        # lp=2 -> threads=2
        # lp=3 -> threads=8
        # lp=4 -> threads=12
        # lp=5 -> threads=16
        # lp=6 -> threads=20
        return {  # threads to lp
            1: 1,
            2: 2, 3: 2, 4: 2, 5: 2,
            6: 3, 7: 3, 8: 3, 9: 3, 10: 3,
            11: 4, 12: 4, 13: 4, 14: 4,
            15: 5, 16: 5, 17: 5,
        }.get(threads, 6)

    if kwargs["mode"] == "vbr":
        return [
            "-c:v", "libsvtav1",
            "-crf", str(round(kwargs["quality"]*63.0)),
            "-preset", {"slow": "4", "medium": "6", "fast": "8"}[kwargs["effort"]],
            # "-tune", "ssim",  # not same result as -svtav1-params tune=2
            "-svtav1-params", f"film-grain=0:lp={libsvtav1_lp(kwargs['threads'])}:tune=2",
        ]
    rate = f"{min(100_000, quality_to_rate(kwargs))}k"
    return [
        "-c:v", "libsvtav1",
        "-b:v", rate, "-minrate", rate, "-bufsize", rate,
        "-preset", {"slow": "4", "medium": "6", "fast": "8"}[kwargs["effort"]],
        "-tune", "ssim",  # -svtav1-params tune=2 not supported in cbr
        "-svtav1-params", f"rc=1:film-grain=0:lp={libsvtav1_lp(kwargs['threads'])}",
    ]


@decorator_pix_fmt
def _encode_libvpx_vp9(**kwargs: dict) -> list[str]:
    """Return the ffmpeg arguments."""
    # https://trac.ffmpeg.org/wiki/Encode/VP9
    # https://wiki.webmproject.org/ffmpeg/vp9-encoding-guide
    # https://developers.google.com/media/vp9/settings
    if kwargs["mode"] == "vbr":
        quality = ["-crf", str(round(kwargs["quality"]*63.0)), "-b:v", "0"]
    else:
        rate = f"{quality_to_rate(kwargs)}k"
        quality = ["-b:v", rate, "-minrate", rate, "-maxrate", rate, "-lag-in-frames", "0"]
    return [
        "-c:v", "libvpx-vp9",
        *quality,
        # in [-16, 16]
        "-speed", {"slow": "-2", "medium": "1", "fast": "8"}[kwargs["effort"]],
        "-tune", "ssim",
        "-row-mt", "0", "-threads", str(kwargs["threads"]),
    ]


@decorator_pix_fmt
def _encode_libx264(**kwargs: dict) -> list[str]:
    """Return the ffmpeg arguments."""
    if kwargs["mode"] == "vbr":
        quality = ["-crf", str(round(kwargs["quality"]*51.0, 1))]
    else:
        rate = f"{quality_to_rate(kwargs)}k"
        quality = [  # https://trac.ffmpeg.org/wiki/Encode/H.264#CBRConstantBitRate
            "-b:v", rate,
            "-minrate", rate,
            "-maxrate", rate,
            "-bufsize", rate,
            "-x264-params", "nal-hrd=cbr",
        ]
    return [  # https://ffmpeg.party/x264/
        "-c:v", "libx264",
        *quality,
        "-preset", {"fast": "veryfast", "medium": "medium", "slow": "veryslow"}[kwargs["effort"]],
        "-tune", "ssim",
        "-threads", str(kwargs["threads"]), "-thread_type", "frame",
    ]


@decorator_pix_fmt
def _encode_libx265(**kwargs: dict) -> list[str]:
    """Return the ffmpeg arguments."""
    if kwargs["mode"] == "vbr":
        return [   # https://x265.readthedocs.io/en/master/cli.html
            "-c:v", "libx265",
            "-crf", str(round(kwargs["quality"]*51.0, 1)),
            "-preset", kwargs["effort"],
            "-tune", "ssim",
            "-x265-params",
            (
                f"frame-threads={kwargs['threads']}:"
                f"pools={kwargs['threads']}:"
                f"wpp={1 if kwargs['threads'] != 1 else 0}"
            ),
        ]
    rate = quality_to_rate(kwargs)
    return [   # https://x265.readthedocs.io/en/master/cli.html
        "-c:v", "libx265",
        "-b:v", f"{rate}k",
        "-preset", {"fast": "veryfast", "medium": "medium", "slow": "veryslow"}[kwargs["effort"]],
        "-tune", "ssim",
        "-x265-params",
        (
            f"vbv-maxrate={rate}:vbv-bufsize={rate}:"
            f"frame-threads={kwargs['threads']}:"
            f"pools={kwargs['threads']}:"
            f"wpp={1 if kwargs['threads'] != 1 else 0}"
        ),
    ]


def _encode_vp9_vaapi(**kwargs: dict) -> list[str]:
    """Return the ffmpeg arguments."""
    return _encode_vaapi("vp9", **kwargs)


@decorator_pix_fmt
def _encode_vvc(**kwargs: dict) -> list[str]:
    """Return the ffmpeg arguments."""
    # https://github.com/fraunhoferhhi/vvenc/wiki/FFmpeg-Integration
    if kwargs["mode"] == "vbr":
        quality = ["-qp", str(round(kwargs["quality"]*63.0))]
    else:
        rate = quality_to_rate(kwargs)
        quality = ["-b:v", f"{rate}k", "-maxrate", f"{round(1.5*rate)+1}k"]
    bit = int(re.search(r"(?P<bit>\d+)le", kwargs["pix_fmt"] + "8le")["bit"])
    return [
        "-c:v", "vvc",
        *quality,
        "-preset", kwargs["effort"],
        "-qpa", "1",
        "-vvenc-params", f"internalbitdepth={bit}",
        "-threads", str(kwargs["threads"]),
    ]



"""Mesures d'Encodage et Décodage Vidéo."""

from .measures import Activity
from .measures.context import full_context as context

__all__ = ["Activity", "context"]
__author__ = "Robin RICHARD (robinechuca)"
__version__ = "1.2.7"  # pep 440



"""Decompress a file."""

import logging
import lzma
import pathlib
import re
import subprocess

import tqdm
from context_verbose import Printer


def _get_size(compressed_file: pathlib.Path) -> int:
    """Try to get the decompressed size with 'xz'."""
    # this method should be improved
    try:
        out = subprocess.run(
            ["xz", "--robot", "--list", str(compressed_file)], check=True, capture_output=True,
        )
    except FileNotFoundError as err:
        logging.getLogger(__name__).warning("please install xz %s", err)
    sizes = re.findall(br"\d+", out.stdout)
    return max(map(int, sizes), default=0)


def decompress(compressed_file: pathlib.Path) -> pathlib.Path:
    """Decompress a *.xz file in the same folder."""
    assert isinstance(compressed_file, pathlib.Path), compressed_file.__class__.__name__
    assert compressed_file.suffix == ".xz", compressed_file
    decompressed_file = compressed_file.parent / compressed_file.stem

    # shortcut
    size = None
    if decompressed_file.exists():
        size = _get_size(compressed_file)
        if decompressed_file.stat().st_size == size:
            return decompressed_file

    # decompress
    with Printer(f"Decompress {compressed_file.name!r}...", color="green") as prt:
        if size is None:
            size = _get_size(compressed_file)

        with (
            tqdm.tqdm(
                dynamic_ncols=True,
                leave=True,
                smoothing=1e-6,
                total=round(size*1e-6, 1),
                unit="Mo",
            ) as load,
            lzma.open(compressed_file, "r") as src,
            decompressed_file.open("wb") as dst,
        ):

            while data := src.read(1_000_000):
                dst.write(data)
                load.total = max(load.total, load.n + len(data)*1e-6)
                load.update(len(data)*1e-6)
            prt.print_time()

    return decompressed_file



"""Perform the metrics and video properties measures."""

import datetime
import logging
import pathlib
import sqlite3

import cutcutcodec
import numpy as np
import orjson
from context_verbose import Printer
from flufl.lock import Lock

from mendevi.database.serialize import list_to_binary
from mendevi.utils import compute_video_hash


def open_ro(database: pathlib.Path, conn: sqlite3.Connection | None) -> sqlite3.Connection:
    """Open a new conn if the previous one is not already opened."""
    if conn is None:
        conn = sqlite3.connect(f"file:{database}?mode=ro", uri=True, timeout=30)
        conn.row_factory = sqlite3.Row
    return conn


def _fill_vid_table(
    database: pathlib.Path,
    vid_id: bytes,
    video: pathlib.Path,
    conn: sqlite3.Connection | None = None,
    **kwargs: dict,
) -> sqlite3.Connection | None:
    """Complete the table t_vid_video."""
    # get actual informations
    conn = open_ro(database, conn)
    prop = conn.execute(
        "SELECT * FROM t_vid_video WHERE vid_id=?", (vid_id,),
    ).fetchone()
    prop = {} if prop is None else dict(prop)

    # missing fields
    fields = {
        "vid_codec",
        "vid_duration",
        "vid_eotf",
        "vid_fps",
        "vid_frames",
        "vid_gamut",
        "vid_height",
        "vid_name",
        "vid_pix_fmt",
        "vid_range",
        "vid_size",
        "vid_width",
        *(("vid_rms_sobel",) if kwargs["rms_sobel"] else ()),
        *(("vid_rms_time_diff",) if kwargs["rms_time_diff"] else ()),
        *(("vid_uvq",) if kwargs["uvq"] else ()),
    }
    fields -= {k for k, v in prop.items() if v is not None}
    if not fields:
        return conn
    conn.close()

    # fill missing fields
    pad = max(map(len, fields)) - 4
    with Printer(f"Get properties of {video.name}...", color="green") as prt:
        # basic fields
        for field in fields - {"vid_rms_sobel", "vid_rms_time_diff", "vid_uvq"}:
            prop[field] = _fill_vid_table_prop(field, prt, pad, video)

        # metric fields
        new_metrics = cutcutcodec.video_metrics(
            video,
            rms_sobel="vid_rms_sobel" in fields,
            rms_time_diff="vid_rms_time_diff" in fields,
            uvq="vid_uvq" in fields,
        )
        for metric in sorted(new_metrics):
            values = new_metrics[metric]
            prt.print(f"{metric:<{pad}}: {np.nanmean(values):.2f} +/- {np.nanstd(values):.2f}")
        prop |= {f"vid_{m}": list_to_binary(v) for m, v in new_metrics.items()}

    # update result
    fields = list(fields)  # to frozen order for sql request
    with (
        Lock(str(database.with_name(".dblock")), lifetime=datetime.timedelta(seconds=600)),
        sqlite3.connect(database) as conn,
    ):
        cursor = conn.cursor()
        cursor.execute(
            "INSERT OR IGNORE INTO t_vid_video (vid_id) VALUES (?)", (vid_id,),
        )
        cursor.execute(
            (
                f"UPDATE t_vid_video SET {', '.join(f'{k}=?' for k in fields)} "
                "WHERE vid_id=?"
            ),
            [prop[k] for k in fields] + [vid_id],
        )
    return None


def _fill_vid_table_prop(field: str, prt: Printer, pad: int, video: pathlib.Path) -> object:
    """Help _fill_vid_table."""
    prop = None
    match field:
        case "vid_duration":
            prop = float(cutcutcodec.get_duration_video(video))
            prt.print(f"{'duration':<{pad}}: ", end="")
            prt.print_time(prop, print_headers=False)
        case "vid_fps":
            prop = float(cutcutcodec.get_rate_video(video))
            prt.print(f"{'fps':<{pad}}: {prop:.2f} Hz")
        case "vid_frames":
            header, info = cutcutcodec.core.analysis.ffprobe.get_slices_metadata(video)
            header, info = header[0], info[0]
            frames = [dict(zip(header, line, strict=False)) for line in info.tolist()]
            prop = orjson.dumps(
                frames, option=orjson.OPT_INDENT_2|orjson.OPT_SORT_KEYS,
            ).decode()
            prt.print(f"{'frames':<{pad}}: {len(frames)} frames")
        case "vid_height":
            prop = cutcutcodec.get_resolution(video)[0]
            prt.print(f"{'height':<{pad}}: {prop} pixels")
        case "vid_width":
            prop = cutcutcodec.get_resolution(video)[1]
            prt.print(f"{'width':<{pad}}: {prop} pixels")
        case "vid_size":
            prop = video.stat().st_size
            prt.print(f"{'size':<{pad}}: {prop*1e-6:.2f} MB")
        case "vid_pix_fmt" | "vid_range" | "vid_name" | "vid_eotf" | "vid_gamut" | "vid_codec":
            prop = {
                "vid_pix_fmt": cutcutcodec.get_pix_fmt,
                "vid_range": cutcutcodec.get_range,
                "vid_name": lambda p: p.name,
                "vid_eotf": lambda p: cutcutcodec.get_colorspace(p).transfer,
                "vid_gamut": lambda p: cutcutcodec.get_colorspace(p).primaries,
                "vid_codec": cutcutcodec.get_codec_video,
            }[field](video)
            prt.print(f"{' '.join(field.split('_')[1:]):<{pad}}: {prop}")
        case _:
            logging.getLogger(__name__).warning("field %s not recognised, skipped", field)
    return prop


def _fill_met_table(
    database: pathlib.Path,
    ref_id: bytes,
    ref_path: pathlib.Path,
    dis_id: bytes,
    dis_path: pathlib.Path,
    **kwargs: dict,
) -> sqlite3.Connection | None:
    """Complete the table t_met_metric."""
    # get actual informations
    conn = open_ro(database, kwargs.pop("conn"))
    metrics = conn.execute(
        "SELECT * FROM t_met_metric WHERE met_ref_vid_id=? AND met_dis_vid_id=?",
        (ref_id, dis_id),
    ).fetchone()
    metrics = {} if metrics is None else dict(metrics)

    # missing fields
    fields = {
        *(("met_lpips_alex",) if kwargs["lpips_alex"] else ()),
        *(("met_lpips_vgg",) if kwargs["lpips_vgg"] else ()),
        *(("met_psnr",) if kwargs["psnr"] else ()),
        *(("met_ssim",) if kwargs["ssim"] else ()),
        *(("met_vif",) if kwargs["vif"] else ()),
        *(("met_vmaf",) if kwargs["vmaf"] else ()),
    }
    fields -= {k for k, v in metrics.items() if v is not None}
    if not fields:
        return conn
    conn.close()
    fields = sorted(fields)  # to frozen order in sql request and for printing repetability

    # fill missing fields
    pad = max(map(len, fields)) - 4
    with Printer(
        f"Compute metrics between {ref_path.name} and {dis_path.name}...", color="green",
    ) as prt:
        new_metrics = cutcutcodec.video_metrics(
            dis_path, ref_path,
            lpips_alex=kwargs["lpips_alex"],
            lpips_vgg=kwargs["lpips_vgg"],
            psnr=kwargs["psnr"],
            ssim=kwargs["ssim"],
            vif=kwargs["vif"],
            vmaf=kwargs["vmaf"],
        )
        for metric in sorted(new_metrics):
            values = new_metrics[metric]
            prt.print(f"{metric:<{pad}}: {np.nanmean(values):.2f} +/- {np.nanstd(values):.2f}")
    metrics = {f"met_{m}": list_to_binary(v) for m, v in new_metrics.items()}

    # update result
    with (
        Lock(str(database.with_name(".dblock")), lifetime=datetime.timedelta(seconds=600)),
        sqlite3.connect(database) as conn,
    ):
        cursor = conn.cursor()
        cursor.execute(
            "INSERT OR IGNORE INTO t_met_metric (met_ref_vid_id, met_dis_vid_id) VALUES (?, ?)",
            (ref_id, dis_id),
        )
        cursor.execute(
            (
                f"UPDATE t_met_metric SET {', '.join(f'{k}=?' for k in fields)} "
                "WHERE met_ref_vid_id=? AND met_dis_vid_id=?"
            ),
            [metrics[k] for k in fields] + [ref_id, dis_id],
        )

    return None


def probe_and_store(
    database: pathlib.Path,
    video: pathlib.Path,
    conn: sqlite3.Connection | None = None,
    **kwargs: dict,
) -> sqlite3.Connection | None:
    """Measure the properties of the video.

    Parameters
    ----------
    database : pathlike
        The path of the existing database to be updated.
    video : pathlib.Path
        The source video file to be annalysed.
    conn: sqlite3.Connection, optional
        A read only database connection.
    **kwargs : dict
        All the metrics

    """
    assert isinstance(video, pathlib.Path), video.__class__.__name__
    assert video.is_file(), video

    vid_id: bytes = compute_video_hash(video)

    conn = _fill_vid_table(database, vid_id, video, conn=conn, **kwargs)
    conn = open_ro(database, conn)

    # get the references videos for comparative comparisons
    references: dict[pathlib.Path, bytes] = kwargs["ref"].copy()
    for res in conn.execute(
        """
        SELECT vid_name, enc_src_vid_id FROM t_enc_encode
        JOIN t_vid_video ON enc_src_vid_id=vid_id
        WHERE enc_dst_vid_id=?
        """,
        (vid_id,),
    ):
        ref_name, ref_id = res["vid_name"], res["enc_src_vid_id"]
        # try to find video full path
        if (
            ref := database.with_name(ref_name)).is_file() or (ref := video.with_name(ref_name)
        ).is_file():
            references[ref] = ref_id
        else:
            logging.getLogger(__name__).info("failed to find the reference video %s", ref)
    references = {ref: ref_id for ref, ref_id in references.items() if ref_id != vid_id}

    # perform the comparative metrics
    for ref, ref_id in references.items():
        conn = _fill_met_table(database, ref_id, ref, vid_id, video, conn=conn, **kwargs)
    return conn



"""Initialize the SQL database."""

import pathlib
import sqlite3

ENV_UNIQUE = [
    "env_ffmpeg_version",
    "env_hostname",
    "env_logical_cores",
    "env_processor",
]

SQL_HEADER_SIZE = 100  # SQLite database file header is 100 bytes


def create_database(filename: str | bytes | pathlib.Path) -> None:
    """Create a new SQL database to store all video informations.

    Parameters
    ----------
    filename : pathlike
        The path of the new database to be created.

    Examples
    --------
    >>> import os, tempfile
    >>> from mendevi.database.create import create_database
    >>> create_database(database := tempfile.mktemp(suffix=".sqlite"))
    >>> os.remove(database)
    >>>

    """
    filename = pathlib.Path(filename).expanduser().resolve()
    assert not filename.exists(), f"the database has to be new, {filename} exists"

    with sqlite3.connect(filename) as conn:
        # next line can fail flush
        # conn.execute("PRAGMA journal_mode=WAL")  # reduce "database is locked"
        cursor = conn.cursor()
        cursor.execute("""CREATE TABLE IF NOT EXISTS t_act_activity (
            act_id INTEGER PRIMARY KEY AUTOINCREMENT,

            /* MEASURES */
            act_start TIMESTAMP NOT NULL,
            act_duration FLOAT NOT NULL CHECK(act_duration > 0.0),
            act_rapl_dt LONGBLOB,
            act_rapl_power LONGBLOB,
            act_wattmeter_dt LONGBLOB,
            act_wattmeter_power LONGBLOB,
            act_gpu_dt LONGBLOB,
            act_gpu_power LONGBLOB,
            act_ps_dt LONGBLOB,
            act_ps_core LONGBLOB,
            act_ps_temp LONGBLOB,
            act_ps_ram LONGBLOB
        )""")
        cursor.execute("""CREATE TABLE IF NOT EXISTS t_vid_video (
            vid_id BINARY(128) PRIMARY KEY,
            vid_name TEXT,

            /* VIDEO CONTENT */
            vid_codec TINYTEXT,
            vid_duration FLOAT CHECK(vid_duration > 0.0),
            vid_eotf TINYTEXT,
            vid_fps FLOAT CHECK(vid_fps > 0.0),
            vid_frames LONGTEXT,
            vid_gamut TINYTEXT,
            vid_height SMALLINT CHECK(vid_height > 0),
            vid_pix_fmt TINYTEXT,
            vid_range TINYTEXT CHECK(vid_range IN ('pc', 'tv')),
            vid_size BIGINT CHECK(vid_width >= 0),
            vid_width SMALLINT CHECK(vid_width > 0),

            /* NON COMPARATIVE METRICS */
            vid_rms_sobel LONGBLOB,
            vid_rms_time_diff LONGBLOB,
            vid_uvq LONGBLOB
        )""")
        cursor.execute("""CREATE TABLE IF NOT EXISTS t_met_metric (
            met_id INTEGER PRIMARY KEY AUTOINCREMENT,
            met_ref_vid_id BINARY(128) NOT NULL,
            met_dis_vid_id BINARY(128) NOT NULL,

            /*  COMPARATIVE METRICS */
            met_lpips_alex LONGBLOB,
            met_lpips_vgg LONGBLOB,
            met_psnr LONGBLOB,
            met_ssim LONGBLOB,
            met_vif LONGBLOB,
            met_vmaf LONGBLOB,

            UNIQUE(met_ref_vid_id, met_dis_vid_id) ON CONFLICT FAIL
        )""")
        cursor.execute(f"""CREATE TABLE IF NOT EXISTS t_env_environment (
            env_id INTEGER PRIMARY KEY AUTOINCREMENT,

            /* CONTEXT DETAILS */
            env_ffmpeg_version MEDIUMTEXT NOT NULL,
            env_hostname TINYTEXT NOT NULL,
            env_kernel_version TINYTEXT,
            env_librav1e_version MEDIUMTEXT,
            env_libsvtav1_version MEDIUMTEXT,
            env_libvpx_vp9_version MEDIUMTEXT,
            env_libx265_version MEDIUMTEXT,
            env_logical_cores INTEGER NOT NULL CHECK(env_logical_cores > 0),
            env_lshw LONGTEXT,
            env_lspci LONGTEXT,
            env_physical_cores INTEGER,
            env_pip_freeze MEDIUMTEXT,
            env_processor TINYTEXT,
            env_python_compiler TINYTEXT,
            env_python_version TINYTEXT,
            env_ram INTEGER NOT NULL CHECK(env_ram > 0),
            env_swap INTEGER,
            env_system_version MEDIUMTEXT,
            env_vvc_version MEDIUMTEXT,

            /* IDLE MEASURES */
            env_idle_act_id INTEGER REFERENCES t_act_activity(act_id),  -- link to activity table

            /* CONSTRAINTS */
            UNIQUE({", ".join(ENV_UNIQUE)}) ON CONFLICT FAIL
        )""")
        cursor.execute("""CREATE TABLE IF NOT EXISTS t_dec_decode (
            dec_id INTEGER PRIMARY KEY AUTOINCREMENT,
            dec_vid_id BINARY(128) NOT NULL REFERENCES t_vid_video(vid_id) ON DELETE CASCADE,
            dec_env_id INTEGER NOT NULL REFERENCES t_env_environment(env_id) ON DELETE CASCADE,
            dec_act_id INTEGER REFERENCES t_act_activity(act_id),
            dec_cmd TEXT,
            dec_decoder TINYTEXT,
            dec_filter TEXT,
            dec_height SMALLINT CHECK(dec_height > 0),
            dec_pix_fmt TINYTEXT,
            dec_width SMALLINT CHECK(dec_width > 0)
        )""")
        cursor.execute("""CREATE TABLE IF NOT EXISTS t_enc_encode (
            enc_id INTEGER PRIMARY KEY AUTOINCREMENT,
            enc_src_vid_id BINARY(128) REFERENCES t_vid_video(vid_id),
            enc_dst_vid_id BINARY(128) NOT NULL REFERENCES t_vid_video(vid_id) ON DELETE CASCADE,
            enc_env_id INTEGER NOT NULL REFERENCES t_env_environment(env_id) ON DELETE CASCADE,
            enc_act_id INTEGER REFERENCES t_act_activity(act_id),

            /* TASK DESCRIPTION */
            enc_cmd TEXT,  -- exact ffmpeg command used
            enc_effort TINYTEXT CHECK(enc_effort IN ('fast', 'medium', 'slow')),
            enc_encoder TINYTEXT,
            enc_filter TEXT,
            enc_fps FLOAT CHECK(enc_fps > 0.0),
            enc_height SMALLINT CHECK(enc_height > 0),
            enc_mode TINYTEXT CHECK(enc_mode IN ('cbr', 'vbr')),
            enc_pix_fmt TINYTEXT,
            enc_quality FLOAT CHECK(enc_quality >= 0.0 AND enc_quality <= 1.0),
            enc_threads SMALLINT CHECK(enc_threads >= 0),
            enc_width SMALLINT CHECK(enc_width > 0)
        )""")
        cursor.execute("""
            CREATE TRIGGER IF NOT EXISTS trigger_insert_env_act_unicity
            BEFORE INSERT ON t_env_environment
            BEGIN
                SELECT
                    CASE
                        WHEN NEW.env_idle_act_id IN (
                            SELECT enc_act_id FROM t_enc_encode
                            UNION ALL SELECT dec_act_id FROM t_dec_decode
                            UNION ALL SELECT env_idle_act_id FROM t_env_environment
                        ) THEN
                            RAISE (ABORT, 'act_id has to be unique')
                    END;
            END;
        """)
        cursor.execute("""
            CREATE TRIGGER IF NOT EXISTS trigger_update_env_act_unicity
            BEFORE UPDATE OF env_idle_act_id ON t_env_environment
            BEGIN
                SELECT
                    CASE
                        WHEN NEW.env_idle_act_id IN (
                            SELECT enc_act_id FROM t_enc_encode
                            UNION ALL SELECT dec_act_id FROM t_dec_decode
                            UNION ALL SELECT env_idle_act_id FROM t_env_environment
                        ) THEN
                            RAISE (ABORT, 'act_id has to be unique')
                    END;
            END;
        """)
        cursor.execute("""
            CREATE TRIGGER IF NOT EXISTS trigger_insert_enc_act_unicity
            BEFORE INSERT ON t_enc_encode
            BEGIN
                SELECT
                    CASE
                        WHEN NEW.enc_act_id IN (
                            SELECT enc_act_id FROM t_enc_encode
                            UNION ALL SELECT dec_act_id FROM t_dec_decode
                            UNION ALL SELECT env_idle_act_id FROM t_env_environment
                        ) THEN
                            RAISE (ABORT, 'act_id has to be unique')
                    END;
            END;
        """)
        cursor.execute("""
            CREATE TRIGGER IF NOT EXISTS trigger_insert_enc_act_unicity
            BEFORE UPDATE OF enc_act_id ON t_enc_encode
            BEGIN
                SELECT
                    CASE
                        WHEN NEW.enc_act_id IN (
                            SELECT enc_act_id FROM t_enc_encode
                            UNION ALL SELECT dec_act_id FROM t_dec_decode
                            UNION ALL SELECT env_idle_act_id FROM t_env_environment
                        ) THEN
                            RAISE (ABORT, 'act_id has to be unique')
                    END;
            END;
        """)
        cursor.execute("""
            CREATE TRIGGER IF NOT EXISTS trigger_insert_dec_act_unicity
            BEFORE INSERT ON t_dec_decode
            BEGIN
                SELECT
                    CASE
                        WHEN NEW.dec_act_id IN (
                            SELECT enc_act_id FROM t_enc_encode
                            UNION ALL SELECT dec_act_id FROM t_dec_decode
                            UNION ALL SELECT env_idle_act_id FROM t_env_environment
                        ) THEN
                            RAISE (ABORT, 'act_id has to be unique')
                    END;
            END;
        """)
        cursor.execute("""
            CREATE TRIGGER IF NOT EXISTS trigger_insert_dec_act_unicity
            BEFORE UPDATE OF dec_act_id ON t_dec_decode
            BEGIN
                SELECT
                    CASE
                        WHEN NEW.dec_act_id IN (
                            SELECT enc_act_id FROM t_enc_encode
                            UNION ALL SELECT dec_act_id FROM t_dec_decode
                            UNION ALL SELECT env_idle_act_id FROM t_env_environment
                        ) THEN
                            RAISE (ABORT, 'act_id has to be unique')
                    END;
            END;
        """)

    filename.chmod(0o777)


def is_sqlite(file: str | bytes | pathlib.Path) -> bool:
    """Test if the provided path is an sqlite3 database.

    Examples
    --------
    >>> import os, pathlib, tempfile
    >>> from mendevi.database.create import create_database, is_sqlite
    >>> database = pathlib.Path(tempfile.mktemp())
    >>> is_sqlite(database)
    False
    >>> create_database(database)
    >>> is_sqlite(database)
    True
    >>> os.remove(database)
    >>>

    """
    file = pathlib.Path(file).expanduser()
    if not file.is_file():
        return False
    with file.open("rb") as raw:
        header = raw.read(SQL_HEADER_SIZE)
    if len(header) < SQL_HEADER_SIZE:
        return False
    return header.startswith(b"SQLite format 3")



"""Defines the different video profiles used.

For YouTube <https://support.google.com/youtube/answer/1722171>:
- muxer mp4
- audio aac stereo 384kbit/s 48kHz
- video h264 yuv420p 1080p 8Mbit/s 29.97fps
<https://trac.ffmpeg.org/wiki/Encode/YouTube>
<https://gist.github.com/mikoim/27e4e0dc64e384adbcb91ff10a2d3678>
"""

from fractions import Fraction

PROFILES = {
    "sd": {
        "fps": Fraction(24000, 1001),
        "pix_fmt": "yuv420p",
        "primaries": "smpte170m",
        "range": "tv",
        "resolution": (480, 720),
        "transfer": "smpte170m",
    },
    "hd": {
        "fps": Fraction(25000, 1001),
        "pix_fmt": "yuv420p",
        "primaries": "bt709",  # sRGB
        "range": "tv",
        "resolution": (720, 1280),
        "transfer": "iec61966-2-1, iec61966_2_1",  # sRGB
    },
    "fhd": {
        "fps": Fraction(30000, 1001),
        "pix_fmt": "yuv420p10le",
        "primaries": "bt709",
        "range": "tv",
        "resolution": (1080, 1920),
        "transfer": "bt709",
    },
    "uhd4k": {  # partialy given by https://trac.ffmpeg.org/wiki/Encode/AV1
        "fps": Fraction(60000, 1001),
        "pix_fmt": "yuv420p10le",
        "primaries": "bt2020",
        "range": "pc",
        "resolution": (2160, 3840),
        "transfer": "smpte2084",
    },
}



"""Query the GPUs activity."""

import numbers
import threading
import time

import numpy as np
import pynvml  # uv pip install nvidia-ml-py

# initialisation
try:
    pynvml.nvmlInit()
except pynvml.NVMLError:
    GPUS = 0
else:
    GPUS: int = pynvml.nvmlDeviceGetCount()


def measure() -> dict[str]:
    """Get a instantaneous capture of all gpus."""
    memory, gpus, powers = [], [], []
    for i in range(GPUS):
        hand = pynvml.nvmlDeviceGetHandleByIndex(i)
        powers.append(pynvml.nvmlDeviceGetPowerUsage(hand) / 1000.0)  # in Watts
        gpus.append(pynvml.nvmlDeviceGetUtilizationRates(hand).gpu / 100.0)  # in [0, 1]
        memory.append(pynvml.nvmlDeviceGetMemoryInfo(hand).used)  # in bytes
    return {"memory": memory, "gpus": gpus, "powers": powers}


class UsageGPU(threading.Thread):
    """Use pynvml through a python context manager.

    Examples
    --------
    >>> import time
    >>> from mendevi.measures.gpu import UsageGPU
    >>> with UsageGPU() as gpu:
    ...     time.sleep(1)
    ...
    >>>

    """

    def __init__(self, sleep: numbers.Real = 50e-3) -> None:
        """Initialize the usage context.

        Parameters
        ----------
        sleep : float, default=50e-3
            The time interval between 2 measures (in s).

        """
        super().__init__(daemon=True)

        assert isinstance(sleep, numbers.Real), sleep.__class__.__name__
        assert sleep > 0, sleep

        self._stop_flag = False
        self.sleep = float(sleep)
        self.res: dict | None = {
            "dt": [],
            "energy": None,
            "gpus": [],
            "memory": [],
            "power": None,
            "powers": [],
        } if GPUS else None

    def run(self) -> None:
        """Perform the measures."""
        while GPUS and not self._stop_flag:
            t_init = time.time()
            for cat, vals in measure().items():
                self.res[cat].append(vals)
            time.sleep(max(0.0, self.sleep + t_init - time.time()))
            self.res["dt"].append(time.time() - t_init)

    def __enter__(self) -> dict:
        """Start to measure.

        Returns
        -------
        Consumption: dict[str]
            * 'dt': The time difference between 2 consecutive power measurements (in s).
            * 'energy': The total energy consumption (in J).
            * 'gpus': The mean usage of all the logical gpus.
            * 'memory': The memory used for each gpu (in bytes).
            * 'power': The average power, energy divided by the duration (in w).
            * 'powers': The power measured between 2 consecutive points (in w).

        """
        self.start()
        return self.res

    def __exit__(self, *_: object) -> None:
        """Stop the measure and update the dictionary returnd by __enter__."""
        self._stop_flag = True
        self.join()  # wait the last update of self.run
        if GPUS:
            self.res["gpu"] = float(np.mean(self.res["gpus"], axis=0).sum())

            # compute total energy
            self.res["energy"] = sum(
                sum(ps) * dt for ps, dt in zip(self.res["powers"], self.res["dt"], strict=False)
            )
            self.res["power"] = self.res["energy"] / sum(self.res["dt"])



#!/usr/bin/env python

"""Command line interface entry point."""

import importlib

import click


class LazyGroup(click.Group):
    """Overwite clik.Group to lasy import."""

    def __init__(self, *args: tuple, **kwargs: dict) -> None:
        """Group initialisation."""
        self._lazy_commands = {}
        super().__init__(*args, **kwargs)

    def add_lazy_command(self, import_path: str, name: str) -> None:
        """Register a command without importing it immediately."""
        self._lazy_commands[name] = import_path

    def get_command(self, ctx: click.Context, cmd_name: str) -> click.Command:
        """Import the function of the module."""
        if cmd_name in self._lazy_commands:
            module_path, func_name = self._lazy_commands[cmd_name].rsplit(".", 1)
            mod = importlib.import_module(module_path)
            return getattr(mod, func_name)
        return super().get_command(ctx, cmd_name)

    def list_commands(self, ctx: click.Context) -> list[str]:
        """List all available command names (including lazy ones)."""
        cmds = set(super().list_commands(ctx))
        cmds.update(self._lazy_commands.keys())
        return sorted(cmds)


@click.group(cls=LazyGroup)
def main() -> None:
    """Perform video transcoding measurements."""


# same as main.add_command but with lazy import
main.add_lazy_command("mendevi.cli.decode.main", "decode")
main.add_lazy_command("mendevi.cli.doc.main", "doc")
main.add_lazy_command("mendevi.cli.download.main", "download")
main.add_lazy_command("mendevi.cli.encode.main", "encode")
main.add_lazy_command("mendevi.cli.g5k.main", "g5k")
main.add_lazy_command("mendevi.cli.json.main", "json")
main.add_lazy_command("mendevi.cli.merge.main", "merge")
main.add_lazy_command("mendevi.cli.plot.main", "plot")
main.add_lazy_command("mendevi.cli.prepare.main", "prepare")
main.add_lazy_command("mendevi.cli.probe.main", "probe")
main.add_lazy_command("mendevi.cli.test.main", "test")
main.add_lazy_command("mendevi.cli.version.main", "version")


if __name__ == "__main__":
    main()



"""Perform decoding measures."""

import contextlib
import datetime
import pathlib
import shlex
import sqlite3
import subprocess

import numpy as np
import orjson
from context_verbose import Printer
from flufl.lock import Lock

from mendevi.database.serialize import list_to_binary, tensor_to_binary
from mendevi.measures import Activity
from mendevi.utils import get_resolution


def decode(vid: pathlib.Path, **kwargs: dict) -> tuple[str, dict[str]]:
    """Decode an existing video.

    Parameters
    ----------
    vid : pathlib.Path
        The source video file to be decoded.
    **kwargs : dict
        Transmitted to :py:func:`get_decode_cmd`.

    Returns
    -------
    cmd : str
        The ffmpeg command.
    activity : dict[str]
        The computeur activity during the decoding process.

    """
    cmd = get_decode_cmd(vid, kwargs.get("filter"), kwargs.get("resolution"))
    prt_cmd = " ".join(map(shlex.quote, [{str(vid): "vid.mp4"}.get(c, c) for c in cmd]))
    with Printer(prt_cmd, color="green") as prt:
        prt.print(f"video: {vid.name}")
        with Activity() as activity:
            subprocess.run(cmd, check=True, capture_output=True)

        # print
        prt.print(f"avg cpu usage: {activity['ps_core']:.1f} %")
        prt.print(f"avg ram usage: {1e-9*np.mean(activity['ps_ram']):.2g} Go")
        if "rapl_power" in activity:
            prt.print(f"avg rapl power: {activity['rapl_power']:.2g} W")
        if "wattmeter_power" in activity:
            prt.print(f"avg wattmeter power: {activity['wattmeter_power']:.2g} W")

    return prt_cmd, activity


def decode_and_store(
    database: pathlib.Path,
    env_id: int,
    vid: pathlib.Path,
    **kwargs: dict,
) -> None:
    """Decode a video file and store the result in the database.

    Parameters
    ----------
    database : pathlike
        The path of the existing database to be updated.
    env_id : int
        The primary integer key of the environment.
    vid : pathlib.Path
        The path of the video to be encoded.
    **kwargs
        Transmitted to :py:func:`decode`.

    """
    # decode the video
    cmd, activity = decode(vid, **kwargs)

    with (
        Lock(str(database.with_name(".dblock")), lifetime=datetime.timedelta(seconds=600)),
        sqlite3.connect(database) as conn,
    ):
        cursor = conn.cursor()

        # fill video table
        with contextlib.suppress(sqlite3.IntegrityError):
            cursor.execute(
                "INSERT INTO t_vid_video (vid_id, vid_name) VALUES (?, ?)",
                (kwargs["dec_vid_id"], vid.name),
            )

        # fill activity table
        activity = {
            "act_duration": activity["duration"],
            "act_gpu_dt": list_to_binary(activity.get("gpu_dt", None)),
            "act_gpu_power": tensor_to_binary(activity.get("gpu_powers", None)),
            "act_ps_core": tensor_to_binary(activity["ps_cores"]),
            "act_ps_dt": list_to_binary(activity["ps_dt"]),
            "act_ps_temp": orjson.dumps(
                activity["ps_temp"], option=orjson.OPT_INDENT_2|orjson.OPT_SORT_KEYS,
            ),
            "act_ps_ram": list_to_binary(activity["ps_ram"]),
            "act_rapl_dt": list_to_binary(activity.get("rapl_dt", None)),
            "act_rapl_power": list_to_binary(activity.get("rapl_powers", None)),
            "act_start": activity["start"],
            "act_wattmeter_dt": list_to_binary(activity.get("wattmeter_dt", None)),
            "act_wattmeter_power": list_to_binary(activity.get("wattmeter_powers", None)),
        }
        keys = list(activity)
        (act_id,) = cursor.execute(
            (
                f"INSERT INTO t_act_activity ({', '.join(keys)}) "
                f"VALUES ({', '.join('?'*len(keys))}) RETURNING act_id"
            ),
            [activity[k] for k in keys],
        ).fetchone()

        # fill decode table
        values = {
            "dec_act_id": act_id,
            "dec_cmd": cmd,
            "dec_env_id": env_id,
            "dec_height": kwargs.get("resolution", (None, None))[0],
            "dec_pix_fmt": "rgb24",
            "dec_vid_id": kwargs["dec_vid_id"],
            "dec_width": kwargs.get("resolution", (None, None))[1],
        }
        keys = list(values)
        cursor.execute(
            f"INSERT INTO t_dec_decode ({', '.join(keys)}) VALUES ({', '.join('?'*len(keys))})",
            [values[k] for k in keys],
        )


def get_decode_cmd(
    video: pathlib.Path,
    additional_filter: str,
    resolution: tuple[int, int] | None,
) -> list[str]:
    """Return the ffmpeg decode cmd.

    Parameters
    ----------
    video : pathlib.Path
        The video to be decoded.
        It is required to know the resolution in order to adapt the filter.
    additional_filter : str
        The additional video filter, (can be an empty string).
    resolution : tuple[int, int], optional
        The new (heigh, width) video shape.

    Returns
    -------
    filter : str
        The full ffmpeg decode bash command arguments.

    Examples
    --------
    >>> import cutcutcodec
    >>> from mendevi.decode import get_decode_cmd
    >>> media = cutcutcodec.utils.get_project_root() / "media" / "video" / "intro.webm"
    >>> get_decode_cmd(media, additional_filter="", resolution=None)  # doctest: +ELLIPSIS
    ['ffmpeg', '-threads', '1', '-i', '.../intro.webm', '-vf', 'format=rgb24', '-f', 'null', '-']
    >>> get_decode_cmd(media, additional_filter="", resolution=(480, 720))
    [..., '-vf', 'scale=h=480:w=720:sws_flags=bicubic,format=rgb24', '-f', 'null', '-']
    >>>

    """
    if resolution is not None:
        assert isinstance(resolution, tuple), resolution.__class__.__name__
        assert len(resolution) == 2, resolution
        assert isinstance(resolution[0], int), resolution
        assert isinstance(resolution[1], int), resolution
        assert (resolution[0], resolution[1]) > (0, 0), resolution

    filters: list[str] = []
    if additional_filter:
        filters.append(additional_filter)
    if resolution is not None and get_resolution(video) != resolution:
        filters.append(f"scale=h={resolution[0]}:w={resolution[1]}:sws_flags=bicubic")
    filters.append("format=rgb24")  # to match monitor pixel conversion
    filters = ",".join(filters)

    # test with "ffmpeg -hwaccels", and see all decoders with "ffmpeg -decoders | grep cuvid"
    # ffmpeg -vsync 0 -hwaccel cuda -hwaccel_output_format cuda -c:v h264_cuvid -i vid.mkv -f null -

    # test if available with "ffmpeg -init_hw_device qsv=hw"
    # test with "ffmpeg -hwaccels", and see all decoders with "ffmpeg -decoders | grep qsv"
    # ffmpeg -hide_banner -hwaccel qsv -hwaccel_output_format qsv
    # -c:v h264_qsv -i vid.mkv -vf hwdownload,format=nv12 -f null -
    # j'ai fait sudo apt install libvpl-tools
    # sudo apt install intel-gpu-tools pour avoir intel_gpu_top
    return ["ffmpeg", "-threads", "1", "-i", str(video), "-vf", filters, "-f", "null", "-"]



"""Parse and verify some user input."""

import pathlib
import re

import click
import context_verbose
import Levenshtein
from context_verbose import Printer

from mendevi.cst.labels import LABELS
from mendevi.database.create import create_database, is_sqlite
from mendevi.database.meta import extract_names
from mendevi.utils import unfold_video_files


class PixelParamType(click.ParamType):
    """Parse the pixel format."""

    name = "pixel"

    def convert(self, value: str, param: click.Option, ctx: click.Context) -> str:
        """Normalize pixel format."""
        value = value.lower()
        if not re.fullmatch(r"yuv4[24][024]p(?:10le|12le)?", value):
            self.fail(f"{value!r} is not a valid pixel format", param, ctx)
        return value


class ResolutionParamType(click.ParamType):
    """Parse the resolution."""

    name = "resolution"

    def convert(self, value: str, param: click.Option, ctx: click.Context) -> tuple[int, int]:
        """Convert a video resolution into tuple."""
        if (match := re.search(
            r"(?P<t1>he)?\D*(?P<d1>\d+)\D+?(?P<t2>w)?\D*(?P<d2>\d+)", value.lower(),
        )) is None:
            self.fail(f"{value!r} is not a valid video shape", param, ctx)
        if match["t1"] == "he" or match["t2"] == "w":
            return (int(match["d1"]), int(match["d2"]))
        return (int(match["d2"]), int(match["d1"]))


def _guess_database(files: tuple[str]) -> str | None:
    """Try to find if one of the file is a database file."""
    candidates: set[str] = set()
    for file in files:
        if is_sqlite(file):
            candidates.add(file)
    if len(candidates) == 1:
        return candidates.pop()
    if len(candidates) > 1:
        msg = f"only one database must be provided, {candidates} are given"
        raise ValueError(msg)
    for file_ in files:
        file = pathlib.Path(file_) / "mendevi.db"
        if is_sqlite(file):
            candidates.add(str(file))
    if len(candidates) == 1:
        return candidates.pop()
    return None


def parse_expr(expr: str | None, prt: Printer, start_msg: str) -> None:
    """Ensure that the expression provided is valid."""
    if expr is None:
        return
    assert isinstance(expr, str), expr.__class__.__name__
    variables = extract_names(expr)  # retrieves the names of variables in the expression
    if excess := variables - set(LABELS):
        def dist(label: list[str], excess: set[str]) -> tuple[int, str]:
            return (
                min(Levenshtein.distance(label, e) for e in excess),
                label,
            )
        msg = (
            f"{', '.join(sorted(excess))} are not recognized variables.\n"
            f"Do you mean {', '.join(sorted(LABELS, key=lambda lab: dist(lab, excess)))}?"
        )
        raise ValueError(msg)
    prt.print(f"{start_msg}: {expr}")


def parse_videos_database(
    prt: context_verbose.Printer,
    videos: tuple[str],
    database: str | None = None,
    *,
    _quiet: bool = False,
) -> tuple[list[pathlib.Path], pathlib.Path]:
    """Find or create the database and extract all the videos.

    Parameters
    ----------
    prt : context_verbose.Printer
        The Printer instance to verbose the process.
    videos : tuple[str]
        The full pseudo pathlike video pointers.
    database : str, optional
        The provided link to the video.

    Returns
    -------
    videos : list[pathlib.Path]
        All the existing unfolded video files.
    database : pathlib.Path
        The existing database path.

    """
    # test if database is provided
    assert isinstance(videos, tuple), videos.__class__.__name__
    assert all(isinstance(v, str) for v in videos), videos
    database = database or _guess_database(videos)

    # videos
    videos = list(unfold_video_files(videos))
    if len(videos) == 1 and not _quiet:
        prt.print(f"video     : {videos[0]}")
    elif len(videos) > 1 and not _quiet:
        prt.print(f"videos    : {len(videos)} files founded")

    # database
    if not database:
        database_candidates = {v.parent for v in videos}
        database_candidates = {f / "mendevi.db" for f in database_candidates}
        if not (database := sorted({d for d in database_candidates if d.is_file()})):
            if len(database_candidates) != 1:  # if no ambiguity, we can create it
                msg = "please provide the database path"
                raise ValueError(msg)
            database = database_candidates
        database = database.pop()
    else:
        database = pathlib.Path(database).expanduser()
        database = database / "mendevi.db" if database.is_dir() else database
    if not database.exists():
        create_database(database)
        if not _quiet:
            prt.print(f"database  : {database} (just created)")
    elif not _quiet:
        prt.print(f"database  : {database}")

    return videos, database



"""Preprocess step."""

import fractions
import pathlib
import shutil

import click
import cutcutcodec
from context_verbose import Printer

from mendevi.cst import PROFILES
from mendevi.utils import best_profile, get_resolution

from .parse import PixelParamType, ResolutionParamType


def _parse_args(prt: Printer, video: pathlib.Path, kwargs: dict) -> None:
    """Verification of the arguments."""
    # profile
    kwargs["profile"] = kwargs.get("profile")
    if kwargs["profile"] is not None:
        assert isinstance(kwargs["profile"], str), kwargs["profile"].__class__.__name__
        kwargs["profile"] = kwargs["profile"].lower()
        assert kwargs["profile"] in PROFILES, (kwargs["profile"], PROFILES)
    else:  # autodetect profile based on the number of pixels
        kwargs["profile"] = best_profile(*get_resolution(video))
    prt.print(f"profile   : {kwargs['profile']}")

    # fps
    kwargs["fps"] = kwargs.get("fps") or PROFILES[kwargs["profile"]]["fps"]
    assert isinstance(kwargs["fps"], fractions.Fraction), kwargs["fps"].__class__.__name__
    assert kwargs["fps"] > 0, kwargs["fps"]
    prt.print(f"fps       : {kwargs['fps']}")

    # resolution
    kwargs["resolution"] = (
        kwargs.get("resolution") or PROFILES[kwargs["profile"]]["resolution"]
    )
    assert isinstance(kwargs["resolution"], tuple), kwargs["resolution"].__class__.__name__
    assert len(kwargs["resolution"]) == 2, kwargs["resolution"]
    assert all(isinstance(s, int) and s > 0 for s in kwargs["resolution"]), kwargs["resolution"]
    prt.print(f"resolution: h={kwargs['resolution'][0]}:w={kwargs['resolution'][1]}")

    # pix_fmt
    kwargs["pix_fmt"] = kwargs.get("pix_fmt") or PROFILES[kwargs["profile"]]["pix_fmt"]
    assert isinstance(kwargs["pix_fmt"], str), kwargs["pix_fmt"].__class__.__name__
    prt.print(f"pix_fmt   : {kwargs['pix_fmt']}")

    # primaries
    kwargs["primaries"] = (
        kwargs.get("primaries") or PROFILES[kwargs["profile"]]["primaries"]
    )
    assert isinstance(kwargs["primaries"], str), kwargs["primaries"].__class__.__name__
    assert kwargs["primaries"] in cutcutcodec.core.colorspace.cst.PRIMARIES, kwargs["primaries"]
    prt.print(f"primaries : {kwargs['primaries']}")

    # transfer
    kwargs["transfer"] = kwargs.get("transfer") or PROFILES[kwargs["profile"]]["transfer"]
    assert isinstance(kwargs["transfer"], str), kwargs["transfer"].__class__.__name__
    assert kwargs["transfer"] in cutcutcodec.core.colorspace.cst.TRC, kwargs["transfer"]
    prt.print(f"transfer  : {kwargs['transfer']}")

    # range
    kwargs["range"] = kwargs.get("range") or PROFILES[kwargs["profile"]]["range"]
    assert isinstance(kwargs["range"], str), kwargs["range"].__class__.__name__
    assert kwargs["range"] in {"tv", "pc"}, kwargs["range"]
    prt.print(f"range     : {kwargs['range']}")


@click.command()
@click.argument("video", type=click.Path())
@click.option(
    "-o", "--output",
    type=click.Path(),
    help="The destination folder.",
)
@click.option(
    "-p", "--profile",
    type=click.Choice(list(PROFILES)),
    help="The video profile.",
)
@click.option(
    "-f", "--fps",
    type=fractions.Fraction,
    help="The framerate of the reference video.",
)
@click.option(
    "-r", "--resolution",
    type=ResolutionParamType(),
    help="The video resolution.",
)
@click.option(
    "--pix_fmt",
    type=PixelParamType(),
    help="The video pixel format.",
)
@click.option(
    "--primaries",
    type=click.Choice(list(cutcutcodec.core.colorspace.cst.PRIMARIES)),
    help="The colorspace.",
)
@click.option(
     "--transfer",
    type=click.Choice(list(cutcutcodec.core.colorspace.cst.TRC)),
    help="The electro-optical transfer function.",
)
@click.option(
    "--range",
    type=click.Choice(["tv", "pc"]),
    help="The pixel coding range.",
)
def main(video: str, output: str | None = None, **kwargs: dict) -> pathlib.Path:
    """Pre-Process the source video before performing the transcoding measurements.

    This step is required to provide the following guarantees:

    \b
    * Ensure a square pixel ratio.
    * Ensure that you have a lossless encoded video with a codec that allows for fast decoding.
    * Ensure high-quality resizing without spectral aliasing and blur.
    * Ensure the correct management of colour space and associated metadata.
    * Ensure that the appearance of the video is not distorted.
    * Ensure that the pixel format is recognised.

    \b
    Parameters
    ----------
    video : pathlike
        The source video to be prepared.
    output : pathlike, optional
        The destination folder to store the reference file.
        If not provided, it is store in the same folder as the input video.
    **kwargs: dict
        Please refer to the detailed arguments below.
    profile : str, default=autodetect
        A value among ``sd``, ``hd``, ``fhd`` or ``uhd4k``.
        This represents a typical video profile, which includes a pixel format,
        and image resolution, a sampling frequency, and a colour space.
        For more details, see :py:module:`mendevi.cst.profiles`.
        By default, it is automatically detected based on the number of pixels of the video.
    fps : fractions.Fraction, optional
        The default value is defined in the profile.
        The exact frame rate of the video. The video is not encoded with a variable frame rate.
    resolution : tuple[int, int], optional
        The default value is defined in the profile.
        The resolution of the output video. The interpolation is high quality,
        ensuring minimal blurring, no spectral overlap and preservation of proportions.
    pix_fmt : str, optional
        The default value is defined in the profile.
        The pixel format of the video.
    primaries : str, optional
        The default value is defined in the profile.
        The colorspace primaries name of the tristimulus.
        The possible values are those supported by *cutcutcodec.core.colorspace.cst.PRIMARIES*.
    transfer : str, optional
        The default value is defined in the profile.
        The electro-optical transfer function
        that allows you to define the light intensity curve of the pixels.
        The possible values are those supported by *cutcutcodec.core.colorspace.cst.TRC*.
    range : str, optional
        The default value is defined in the profile.
        You can specify ``tv`` for limited range or ``pc`` for full range.

    """
    with Printer("Parse configuration...") as prt:
        # video
        video = pathlib.Path(video).expanduser().resolve()
        assert video.is_file(), video
        prt.print(f"src video : {video}")

        # output
        if output is None:
            output = video.parent
        else:
            output = pathlib.Path(output).expanduser()
            output.mkdir(mode=0o777, parents=True, exist_ok=True)

        # others
        _parse_args(prt, video, kwargs)

        # find filename
        file = []
        for profile, values in PROFILES.items():
            file.append(["reference", video.stem])
            if kwargs["fps"] != values["fps"]:
                file[-1].append(f"{float(kwargs['fps']):.2f}fps")
            if kwargs["resolution"] != values["resolution"]:
                file[-1].append(f"{kwargs['resolution'][1]}x{kwargs['resolution'][0]}")
            if kwargs["pix_fmt"] != values["pix_fmt"]:
                file[-1].append(kwargs["pix_fmt"])
            if kwargs["primaries"] != values["primaries"]:
                file[-1].append(f"p_{kwargs['primaries']}")
            if kwargs["transfer"] != values["transfer"]:
                file[-1].append(f"t_{kwargs['transfer']}")
            if kwargs["range"] != values["range"]:
                file[-1].append(f"r_{kwargs['range']}")
            file[-1].append(profile)
            file[-1] = "_".join(file[-1]) + ".mp4"
        file = {len(f): f for f in file}
        file = file[min(file)]  # select the shortest name
        file = output / file
        prt.print(f"dst video : {file}")

    # write the file
    if not file.exists():
        tmp = file.with_stem(f"{file.stem}_in_process")
        with cutcutcodec.read(video) as container:
            stream = container.out_select("video")[0]
            cutcutcodec.write(
                [stream],
                tmp,
                colorspace=cutcutcodec.Colorspace(
                    "y'pbpr", kwargs["primaries"], kwargs["transfer"],
                ),
                # streams_settings=[{  # extension .mp4
                #     "encodec": "ffv1",  # fast lossless compression
                #     "rate": kwargs["fps"],
                #     "shape": kwargs["resolution"],
                #     "pix_fmt": kwargs["pix_fmt"],
                #     "range": kwargs["range"],
                # }],

                # https://trac.ffmpeg.org/wiki/Encode/H.264
                streams_settings=[{  # extension .mp4
                    "encodec": "libx264",
                    "options": {"preset": "veryfast", "tune": "fastdecode", "qp": "0"},
                    "rate": kwargs["fps"],
                    "shape": kwargs["resolution"],
                    "pix_fmt": kwargs["pix_fmt"],
                    "range": kwargs["range"],
                }],

                # # https://trac.ffmpeg.org/wiki/Encode/AV1#CRF
                # streams_settings=[{  # extension .mp4
                #     "encodec": "libsvtav1",
                #     "options": {"preset": "4", "svtav1-params": "lossless=1"},
                #     "rate": kwargs["fps"],
                #     "shape": kwargs["resolution"],
                #     "pix_fmt": kwargs["pix_fmt"],
                #     "range": kwargs["range"],
                # }],

                # # https://trac.ffmpeg.org/wiki/Encode/AV1#Losslessencoding
                # streams_settings=[{  # extension .mp4
                #     "encodec": "libaom-av1",
                #     "options": {"crf": "0"},
                #     "rate": kwargs["fps"],
                #     "shape": kwargs["resolution"],
                #     "pix_fmt": kwargs["pix_fmt"],
                #     "range": kwargs["range"],
                # }],

            )
        shutil.move(tmp, file)
        file.chmod(0o777)
    return file



"""Few tools to help display."""

import functools
import numbers

from mendevi.cst.profiles import PROFILES


@functools.cache
def sorted_profiles() -> dict[str, int]:
    """Sort the profiles."""
    profiles = sorted(
        PROFILES, key=lambda p: PROFILES[p]["resolution"][0]*PROFILES[p]["resolution"][1],
    )
    return dict(zip(profiles, range(len(profiles)), strict=True))


def smartsort(item: object) -> object:
    """Make the order relationship more relevant."""
    if isinstance(item, str):
        profiles = sorted_profiles()
        efforts = {"fast": 0, "medium": 1, "slow": 2}  # optional for now because already ordered
        return {**profiles, **efforts}.get(item, item)
    if isinstance(item, numbers.Real):
        return item
    return str(item)



"""Provide simple tools."""

import base64
import functools
import hashlib
import logging
import math
import multiprocessing.pool
import numbers
import pathlib
import re
import typing
from fractions import Fraction

import cutcutcodec
import tqdm
from cutcutcodec.core.io import VIDEO_SUFFIXES

from mendevi.cst.profiles import PROFILES

HASHSIZE = 16  # md5 has number of bytes
PATHLIKE = str | bytes | pathlib.Path


@functools.cache
def best_profile(height: numbers.Integral, width: numbers.Integral) -> str:
    """Return the closest profile name.

    Examples
    --------
    >>> from mendevi.utils import best_profile
    >>> best_profile(1080, 1920)
    'fhd'
    >>>

    """
    assert isinstance(height, numbers.Integral), height.__class__.__name__
    assert isinstance(width, numbers.Integral), width.__class__.__name__
    size = math.sqrt(float(height * width))
    dist_to_profile = {
        abs(math.sqrt(float(v["resolution"][0]*v["resolution"][1])) - size): p
        for p, v in PROFILES.items()
    }
    return dist_to_profile[min(dist_to_profile)]


def compute_video_hash(
    videos: PATHLIKE | typing.Iterable[PATHLIKE], *, fast: bool=True,
) -> bytes | dict[pathlib.Path, bytes]:
    r"""Compute the checksum of the video.

    For :math:`n` hash of :math:`b` bits, the proba of the colision :math:`C` is
    :math:`p(C) = 1 - \left(\frac{2^k-1}{2^k}\right)^{\frac{n(n-1)}{2}}`.

    The md5 hash uses :math:`b = 128` bits. If we add one video per second durring 10 years,
    the proba of colision is about :math:`p(C) \approx 1.46*10^{-22}`.

    That's why the md5 hash is used to identify the video files.

    Parameters
    ----------
    videos : pathlike or list[pathlike]
        The single or set of video you want to compute the signature.
    fast : boolean, default=True
        If the hash appears in the file name, it is returned immediately.
        If False, the actual checksum is systematically recalculated.

    Returns
    -------
    signatures
        The md5 checksum of the video file. In the case of a multiple file,
        a dictionary containing the file and the hash is returned rather a single hash.
        If the file does not exists, return None.

    """
    assert isinstance(fast, bool), fast.__class__.__name__

    def _hash(video_fast: tuple[PATHLIKE, bool]) -> pathlib.Path:
        video, fast = video_fast
        video = pathlib.Path(video)
        # [2-7a-z]{26} for mendevi > 1.2.2
        if fast and (match := re.search(r"[0-9a-f]{32}|[2-7a-z]{26}", video.stem)):
            return video, signature_to_hash(match.group())
        if not (video := video.expanduser()).is_file():
            return video, None
        with video.open("rb") as raw:
            return video, hashlib.file_digest(raw, "md5").digest()

    if isinstance(videos, list | tuple | set | frozenset):
        with multiprocessing.pool.ThreadPool() as pool:
            return dict(tqdm.tqdm(
                pool.imap_unordered(_hash, ((v, fast) for v in videos)),
                desc="compute videos checksum",
                dynamic_ncols=True,
                leave=False,
                smoothing=1e-6,
                total=len(videos),
                unit="video",
            ))
    return _hash((videos, fast))[1]


@functools.cache
def get_pix_fmt(*args: tuple) -> str:
    """Alias to cutcutcodec func."""
    return cutcutcodec.get_pix_fmt(*args)


def get_project_root() -> pathlib.Path:
    """Return the absolute project root folder.

    Examples
    --------
    >>> from mendevi.utils import get_project_root
    >>> root = get_project_root()
    >>> root.is_dir()
    True
    >>> root.name
    'mendevi'
    >>> sorted(p.name for p in root.iterdir())  # doctest: +ELLIPSIS
    ['__init__.py', '__main__.py', ...]
    >>>

    """
    return pathlib.Path(__file__).resolve().parent


@functools.cache
def get_rate_video(*args: tuple) -> Fraction:
    """Alias to cutcutcodec func."""
    return cutcutcodec.get_rate_video(*args)


@functools.cache
def get_resolution(*args: tuple) -> tuple[int, int]:
    """Alias to cutcutcodec func."""
    return cutcutcodec.get_resolution(*args)


def hash_to_signature(checksum: bytes) -> str:
    r"""Convert the md5 binary hash value into an urlsafe string.

    Bijection of :py:func:`signature_to_hash`.

    Parameters
    ----------
    checksum : bytes
        The 128 bit binary hash value.

    Returns
    -------
    signature : str
        The 26 ascii [2-7a-z] symbols string of the converted checksum.

    Examples
    --------
    >>> from mendevi.utils import hash_to_signature
    >>> hash_to_signature(b"\xd4\x1d\x8c\xd9\x8f\x00\xb2\x04\xe9\x80\t\x98\xec\xf8B~")
    'd41d8cd98f00b204e9800998ecf8427e'
    >>>

    """
    assert isinstance(checksum, bytes), checksum.__class__.__name__
    assert len(checksum) == HASHSIZE, len(checksum)
    return checksum.hex()


def signature_to_hash(signature: str) -> bytes:
    r"""Convert the string signature into the md5 checksum.

    Bijection of :py:func:`hash_to_signature`.

    Parameters
    ----------
    signature : str
        The 26 ascii [2-7a-z] symbols string of the converted checksum.

    Returns
    -------
    checksum : bytes
        The 128 bit binary hash value.

    Examples
    --------
    >>> from mendevi.utils import signature_to_hash
    >>> signature_to_hash("d41d8cd98f00b204e9800998ecf8427e")
    b'\xd4\x1d\x8c\xd9\x8f\x00\xb2\x04\xe9\x80\t\x98\xec\xf8B~'
    >>> signature_to_hash("2qoyzwmpaczaj2mabgmoz6ccpy")  # mendevi < 1.2.2
    b'\xd4\x1d\x8c\xd9\x8f\x00\xb2\x04\xe9\x80\t\x98\xec\xf8B~'
    >>>

    """
    assert isinstance(signature, str), signature.__class__.__name__
    try:  # version >= 1.2.2
        assert re.fullmatch(r"[0-9a-f]{32}", signature), signature
    except AssertionError:  # retrocompatibility version < 1.2.2
        assert re.fullmatch(r"[2-7a-z]{26}", signature), signature
        return base64.b32decode(f"{signature.upper()}======".encode())
    else:
        return bytes.fromhex(signature)


def unfold_video_files(
    paths: typing.Iterable[PATHLIKE],
) -> typing.Iterable[pathlib.Path]:
    """Explore recursively the folders to find the video path.

    Parameters
    ----------
    paths : list[pathlike]
        All the folders, files, glob or recursive glob expression.

    Yields
    ------
    filename : pathlib.Path
        The path of the video.

    """
    assert hasattr(paths, "__iter__"), paths.__class__.__name__
    for path_general in paths:
        path = pathlib.Path(path_general).expanduser()
        if path.is_file():
            yield path
        elif path.is_dir():
            for root, _, files in path.walk():
                for file_name in files:
                    file = root / file_name
                    if file.suffix.lower() in VIDEO_SUFFIXES:
                        yield file
        elif "*" in path.name and path.parent.is_dir():
            yield from unfold_video_files(path.parent.glob(path.name))
        elif "**" in (parts := path.parts):
            idx = parts.index("**")
            yield from unfold_video_files(
                pathlib.Path(*parts[:idx]).glob(pathlib.Path(*parts[idx:])),
            )
        else:
            logging.getLogger(__name__).warning("the path %s is not correct", path)



"""Executes analyser of code quality."""

import os
import subprocess

from mendevi.utils import get_project_root


def run_ruff(args: list[str]) -> bool:
    """Run Ruff with given arguments, return True if no lint errors."""
    cmd = ["ruff", "check", *args]
    env = os.environ | {"FORCE_COLOR": "1"}
    result = subprocess.run(cmd, check=False, capture_output=True, text=True, env=env)
    print(result.stdout)
    print(result.stderr)
    return result.returncode == 0


def test_linting() -> None:
    """Run Ruff (replaces mccabe, pycodestyle, pydocstyle, pyflakes, pylint)."""
    root = get_project_root()
    assert run_ruff([str(root), "--config", str(root.parent / "pyproject.toml")])



"""Integrates grid5000's power meter."""

import datetime
import logging
import numbers
import re
import threading

import numpy as np
import requests

GOOD_REQ_STATUS = 200


class G5kPower(threading.Thread):
    """Asynchronous consumption request on g5k."""

    def __init__(self, *args: tuple, **kwargs: dict) -> None:
        """Initialise the power thread."""
        super().__init__(daemon=True)
        self.args = args
        self.kwargs = kwargs
        self.power = None

    def run(self) -> None:
        """Perform the request."""
        try:
            self.power = g5kpower(*self.args, **self.kwargs)
        except ValueError as err:
            logging.getLogger(__name__).warning(err)

    def get(self) -> dict[str]:
        """Retrive the result."""
        self.join()
        return self.power


def g5kpower(
    hostname: str,
    start: numbers.Real,
    duration: numbers.Real,
    *,
    login: str | None = None,
    password: str | None = None,
) -> dict[str]:
    """Do a request to get the grid5000 consumption.

    Parameters
    ----------
    hostname : str
        The hostname containing the node name and the site.
        It can be get with `platform.node()`.
    start : float
        The starting timestamp, it can be get by `time.time()`.
    duration : float
        The job duration in seconds.
    login, password : str
        Username an password for grid5000 api.

    Returns
    -------
    Consumption: dict[str]
        * 'dt': The time difference between 2 consecutive power measurements (in s).
        * 'energy': The total energy consumption (in J).
        * 'power': The average power, energy divided by the duration (in w).
        * 'powers': The power measured between 2 consecutive points (in w).

    Raises
    ------
    ValueError
        If the request failed.

    Examples
    --------
    >>> import platform, time
    >>> from mendevi.measures.g5kpower import g5kpower
    >>> try:
    ...     power = g5kpower(platform.node(), time.time()-10.0, 10.0)
    ... except ValueError:
    ...     pass
    ...
    >>>

    Notes
    -----
    * Tested with `oarsub -I -p paradoxe -t deploy -t monitor='wattmetre_power_watt'`.
    * Power and time measurements are differental to increase the compressibility
      of the database containing this result.
    * Energy is estimated from the trapezoidal integral of instantaneous powers.

    """
    # verification
    assert isinstance(hostname, str), hostname.__class__.__name__
    assert isinstance(start, numbers.Real), start.__class__.__name__
    assert isinstance(duration, numbers.Real), duration.__class__.__name__
    if (
        hostname_fields := re.search(
            r"^(?P<node>[a-z0-9_-]+)\.(?P<site>[a-z0-9_-]+)", hostname, re.IGNORECASE,
        )
    ) is None:
        msg = f"the hostname {hostname} is not grid5000 formated"
        raise ValueError(msg)

    def timestamp_to_str(timestamp: float) -> str:
        return datetime.datetime.fromtimestamp(
            timestamp, tz=datetime.UTC,
        ).strftime("%Y-%m-%dT%H:%M:%S.%f")

    # grid5000 api request
    url = (  # tz=datetime.UTC add a non supported +00:00 by the g5k api
        f"https://api.grid5000.fr/stable/sites/{hostname_fields['site']}/metrics?"
        f"nodes={hostname_fields['node']}&metrics=wattmetre_power_watt"
        f"&start_time={timestamp_to_str(start)}"
        f"&end_time={timestamp_to_str(start+duration)}"
    )
    auth = requests.auth.HTTPBasicAuth(login, password) if login and password else None
    try:
        req = requests.get(url, auth=auth, verify=True, timeout=60)
    except requests.exceptions.SSLError as err:
        logging.getLogger(__name__).warning(err)
        req = requests.get(url, auth=auth, verify=False, timeout=60)
    if req.status_code != GOOD_REQ_STATUS:
        msg = f"the request {url} failed"
        raise ValueError(msg, req)
    if not (req := req.json()):
        msg = f"the request {url} gives an empty result"
        raise ValueError(msg)

    # parse result
    req = {datetime.datetime.fromisoformat(d["timestamp"]).timestamp(): d["value"] for d in req}
    req = {t: v for t, v in req.items() if start <= t <= start + duration}
    times, powers = zip(*req.items(), strict=False)
    times, powers = np.array(times, dtype=np.float64), np.array(powers, dtype=np.float64)

    # pad for accurate boundaries
    order = np.argsort(times)
    times, powers = times[order], powers[order]
    times = np.concatenate([(start,), times, (start+duration,)])  # accurate boundaries
    powers = np.pad(powers, (1, 1), mode="edge")

    # differencial
    power_dt = times[1:] - times[:-1]

    # compute energy, trapeze integral of power
    energy = 0.5 * float(np.sum((powers[:-1] + powers[1:]) * power_dt))

    return {
        "dt": power_dt.tolist(),
        "energy": energy,
        "power": energy / float(times[-1] - times[0]),
        "powers": powers.tolist(),
    }



"""Compile and show the documentation."""

import subprocess
import sys

import click

from mendevi.utils import get_project_root


@click.command()
@click.option("-c", "--recompile", is_flag=True, help="Force to recompile the documentation.")
@click.option("--no-open", is_flag=True, help="Do not open the documentation in the browser.")
def main(*, recompile: bool = False, no_open: bool = False) -> None:
    """Compile and show the documentation."""
    assert isinstance(recompile, bool), recompile.__class__.__name__
    assert isinstance(no_open, bool), no_open.__class__.__name__
    root = get_project_root().parent / "docs"
    index = (root / "build" / "html" / "index.html").resolve()
    if not index.exists() or recompile:
        subprocess.run(
            ["make", "clean"],
            check=True,
            cwd=root,
            stdout=sys.stderr,
        )
        subprocess.run(
            ["make", "html"],
            check=True,
            cwd=root,
            stdout=sys.stderr,
        )
    click.echo(index)
    if not no_open:
        subprocess.run(
            ["xdg-open", str(index)],  # default browser
            check=False,
            stdout=sys.stderr,
        )



"""Power prediction based on CPU utilisation rate."""

import logging
import math

import numpy as np
from context_verbose import Printer

from mendevi.models import Model

MIN_POINTS = 2  # minimal number of points


class PowerCores(Model):
    r"""Affine model to predict the power from the core utilisation.

    .. math:

        P(c) = P_{static} + c \times P_{core}

    With :math:`P_{static}` and :math:`P_{core}`
    two hyperparameters specific to each machine,
    but independent of the programme being executed.

    Examples
    --------
    >>> from mendevi.models.power_cores import PowerCores
    >>> model = PowerCores()
    >>> model.fit("<multithread.db>")
    >>> model.predict(["paradoxe-32.rennes.grid5000.fr"], [0.0])
    >>>

    """

    def __init__(self) -> None:
        """Initialise the model."""
        super().__init__(
            "Power estimation based on the utilisation rate of logic cores",
            sources="""
            F. C. Heinrich, T. Cornebize, A. Degomme, A. Legrand, A. Carpen-
            Amarie, S. Hunold, A.-C. Orgerie, and M. Quinson, “Predicting the
            Energy Consumption of MPI Applications at Scale Using a Single
            Node,” in IEEE Cluster Conference, 2017.

            R. Rodriguez-Sánchez, F. D. Igual, J. L. Martinez, R. Mayo, and E. S.
            Quintana-Orti, “Parallel performance and energy efficiency of modern
            video encoders on multithreaded architectures,” in Eur
            """,
            input_labels=["hostname", "cores"],
            output_labels=["power"],
            parameters={},  # to each hostname, associate P_{static} and P_{core}
        )

    def _fit(self, values: dict[str]) -> None:
        """Perform a linear regression for the 2 constants P_{static} and P_{core}."""
        for hostname in set(values["hostname"]):
            with Printer(f"Affine regression of power = f(cores) on {hostname}...") as prt:
                mask = [i for i, h in enumerate(values["hostname"]) if h == hostname]
                prt.print(f"fit on {len(mask)} points")
                if len(mask) < MIN_POINTS:
                    logging.getLogger(__name__).error(
                        "can not fit model for hostname=%s, not enouth points", hostname,
                    )
                    continue
                cores = [values["cores"][i] for i in mask]
                power = [values["power"][i] for i in mask]
                p_cores, p_static = np.polyfit(cores, power, deg=1)
                prt.print(f"power = {p_static:.3g} + cores * {p_cores:.3g}")
                self.parameters[hostname] = {"p_static": float(p_static), "p_cores": float(p_cores)}

    def _predict(self, values: dict[str]) -> dict[str]:
        all_power = []
        for hostname, cores in zip(values["hostname"], values["cores"], strict=True):
            if hostname not in self.parameters:
                logging.getLogger(__name__).error(
                    "fit the model for the hostname %s before to predict", hostname,
                )
                all_power.append(math.nan)
            else:
                all_power.append(
                    self.parameters[hostname]["p_static"]
                    + cores * self.parameters[hostname]["p_cores"],
                )
        return {"power": all_power}



"""Search for available torrent files."""

import pathlib
import shutil

import requests
from context_verbose import Printer

from mendevi.utils import get_project_root


def get_torrent(name: str) -> pathlib.Path:
    """Retrieve a torrent from its full name.

    Parameters
    ----------
    name : str
        The full torrent name, for example "multithread.db.xz.torrent".

    Returns
    -------
    path : pathlib.Path
        The full path of the torrent.

    Raises
    ------
    KeyError
        If the requested torrent is not included in the list of possible torrents.

    """
    assert isinstance(name, str), name.__class__.__name__

    # look in cachedir
    cachedir = pathlib.Path.home() / ".cache" / "mendevi"
    cachedir.mkdir(mode=0o777, parents=True, exist_ok=True)
    if (file := cachedir / name).exists():
        return file

    # look on the local cloned mndevi git
    local = {p.name: p for p in probe_local_torrent()}
    if name in local:
        # copy in mendevi cachedir for transmission permission
        shutil.copy(local[name], cachedir / name)
        return cachedir / name

    # download online
    with Printer(f"Download {name!r}...", color="green") as prt:
        # look avalable list
        prt.print("get url")
        online = probe_online_torrent()
        if name not in online:
            msg = f"{name!r} not in {', '.join(sorted(set(local) | set(online)))}"
            raise KeyError(msg)

        # download torrent file
        prt.print(f"download {online[name]}")
        req = requests.get(online[name], stream=True, timeout=60)
        req.raise_for_status()
        torrent_data = req.raw.data
        prt.print(f"{len(torrent_data)} bytes retrieved")
        assert torrent_data

    # write the file
    with file.open("wb") as raw:
        raw.write(torrent_data)
    return file


def probe_online_torrent() -> dict[str, str]:
    """Search on GitLab online for the names of available torrents.

    Returns
    -------
    torrents : dict
        For each torrent name, provide the URL for downloading it.

    Examples
    --------
    >>> from mendevi.download.torrent_finder import probe_online_torrent
    >>> sorted(probe_online_torrent())
    ['duration.db.xz.torrent', 'multithread.db.xz.torrent', 'x264_vs_openh264.db.xz.torrent']
    >>>

    """
    url = "https://gitlab.inria.fr/api/v4/projects/rrichard%2Fmendevi/repository/tree?path=dataset"
    req = requests.get(url, timeout=60)
    req.raise_for_status()
    all_files = req.json()
    return {
        f["name"]: f"https://gitlab.inria.fr/rrichard/mendevi/-/raw/main/{f['path']}"
        for f in all_files if f["name"].endswith(".torrent")
    }


def probe_local_torrent() -> set[pathlib.Path]:
    """Search for the names of locally accessible torrents (if the mendevi repository is cloned).

    Returns
    -------
    torrents : set[pathlib.Path]
        Provide all the local torrent files.

    Examples
    --------
    >>> from mendevi.download.torrent_finder import probe_local_torrent
    >>> sorted(t.name for t in probe_local_torrent())
    ['duration.db.xz.torrent', 'multithread.db.xz.torrent', 'x264_vs_openh264.db.xz.torrent']
    >>>

    """
    return set((get_project_root().parent / "dataset").glob("*.torrent"))



"""Test if the dependents libraries are well installed and linked.

Basicaly, it checks if the installation seems to be correct.
"""

import logging
import re
import subprocess

import av
import torch

VERSION_MIN = 6
VERSION_MAX = 8


def test_av_ffmpeg_link() -> None:
    """Test the pyav and ffmpeg dependencies."""
    try:
        version_ffmpeg = (
            subprocess.run(["ffmpeg", "-version"], capture_output=True, check=True).stdout.decode()
        )
    except subprocess.CalledProcessError as err:
        msg = (
            "ffmpeg is not accessible from the 'ffmpeg' commande line, "
            "please refer to the ffmpeg installation guide"
        )
        raise ImportError(
            msg,
        ) from err
    try:
        version_ffprobe = (
            subprocess.run(["ffprobe", "-version"], capture_output=True, check=True).stdout.decode()
        )
    except subprocess.CalledProcessError as err:
        msg = (
            "ffprobe is not accessible from the 'ffprobe' but ffmpeg is installed, "
            "please check your environement variables"
        )
        raise ImportError(
            msg,
        ) from err

    version_pattern = r"version\s+[^\n]*?(?P<major>\d+)\.(?P<minor>\d+)"
    if (version := re.search(version_pattern, version_ffmpeg)) is None:
        msg = "failed to decode the version of ffmpeg"
        raise RuntimeError(msg)
    assert VERSION_MIN <= int(version["major"]) <= VERSION_MAX, (
        "only the version 4, 5, and 6 of ffmpeg are supported, "
        f"not {version['major']}.{version['minor']}"
    )
    if (version := re.search(version_pattern, version_ffprobe)) is None:
        msg = "failed to decode the version of ffprobe"
        raise RuntimeError(msg)
    assert VERSION_MIN <= int(version["major"]) <= VERSION_MAX, (
        "only the version 4, 5, and 6 of ffprobe are supported, "
        f"not {version['major']}.{version['minor']}"
    )

    version_pattern = r"(?P<major>\d+)\.\s*(?P<minor>\d+)\.\s*(?P<micro>\d+)"
    for lib, (major, minor, micro) in av.library_versions.items():
        if (version := re.search(rf"{lib}\s+{version_pattern}", version_ffmpeg)) is None:
            msg = f"failed to extract the ffmpeg {lib} version"
            raise RuntimeError(msg)
        if (
            (major, minor, micro)
            != (int(version["major"]), int(version["minor"]), int(version["micro"]))
        ):
            logging.getLogger(__name__).warning(
                "the av version of %s is %d.%d.%d but %s.%s.%s for ffmpeg, "
                "please reinstall pyav using option --no-binary, see the installation guide "
                "https://pyav.org/docs/develop/overview/installation.html",
                lib, major, minor, micro, version["major"], version["minor"], version["micro"],
            )


def test_gpu_torch() -> None:
    """Test if torch is able to use the GPU."""
    # possible to test lspci | grep ' NVIDIA '
    if torch.cuda.is_available():
        return  # case always ok
    try:
        result = subprocess.run(["lshw", "-C", "display"], capture_output=True, check=True)
    except FileNotFoundError as err:
        try:
            subprocess.run(["nvidia-smi"], capture_output=True, check=True)
        except FileNotFoundError:
            return  # assume there are not graphical card
        msg = (
            "There seems to be an nvidia gpu on this machine, "
            "however torch is not able to use it, please reinstall cuda"
        )
        raise ImportError(msg) from err
    if b" nvidia " in result.stdout.lower():
        msg = (
            "There seems to be an nvidia gpu on this machine, "
            "however torch is not able to use it, please reinstall cuda"
        )
        raise ImportError(msg)



"""Download and decapsule a file."""


import click
from context_verbose import Printer

from mendevi.download.decapsulation import retrive_file


@click.command()
@click.argument("file", type=str)
def main(file: str) -> None:
    """Download and decapsule a file.

    \b
    Parameters
    ----------
    file : pathlike
        The filename to get.

    """
    with Printer(f"Get {file!r}...", color="cyan") as prt:
        path = retrive_file(file)
        prt.print(f"final file: {path}")



"""Get ffmpeg filter chain command."""

import fractions
import pathlib
import re

from mendevi.utils import get_pix_fmt, get_rate_video, get_resolution


def get_convert_cmd(
    video: pathlib.Path | str,
    additional_filter: str,
    fps: fractions.Fraction | None,
    pix_fmt: str | None,
    resolution: tuple[int, int] | None,
) -> str:
    """Generate the ffmpeg filter command that performs this conversion.

    Parameters
    ----------
    video : pathlike
        The video to be filtered.
        It is required to know the relevant filters and find the best possible order.
    additional_filter : str
        The additional video filter, (can be an empty string).
    fps : fractions.Fraction, optional
        The new framerate.
    pix_fmt : str, optional
        The new pixel format, it has to match the regex ``yuv4[24][024]p(?:10le|12le)?``.
    resolution : tuple[int, int], optional
        The new (heigh, width) video shape.

    Returns
    -------
    filter : str
        The ffmpeg video filter argument. It can be an empty string if nothing has to be done.

    Examples
    --------
    >>> from fractions import Fraction
    >>> import cutcutcodec
    >>> from mendevi.convert import get_convert_cmd
    >>> media = cutcutcodec.utils.get_project_root() / "media" / "video" / "intro.webm"
    >>> get_convert_cmd(media, additional_filter="", fps=None, pix_fmt=None, resolution=None)
    ''
    >>> get_convert_cmd(
    ...     media,
    ...     additional_filter="",
    ...     fps=Fraction(60000, 1001),
    ...     pix_fmt="yuv420p10le",
    ...     resolution=(1080, 1920),
    ... )
    'scale=h=1080:w=1920:sws_flags=bicubic,format=yuv420p10le,fps=60000/1001'
    >>> get_convert_cmd(
    ...     media,
    ...     additional_filter="",
    ...     fps=Fraction(24000, 1001),
    ...     pix_fmt="yuv420p",
    ...     resolution=(480, 720),
    ... )
    'fps=24000/1001,scale=h=480:w=720:sws_flags=bicubic'
    >>>

    """
    if fps is not None:
        assert isinstance(fps, fractions.Fraction), fps.__class__.__name__
        assert fps > 0, fps
    if pix_fmt is not None:
        assert isinstance(pix_fmt, str), pix_fmt.__class__.__name__
        assert re.fullmatch(r"yuv4[24][024]p(?:10le|12le)?", pix_fmt), pix_fmt
    if resolution is not None:
        assert isinstance(resolution, tuple), resolution.__class__.__name__
        assert len(resolution) == 2, resolution
        assert isinstance(resolution[0], int), resolution
        assert isinstance(resolution[1], int), resolution
        assert (resolution[0], resolution[1]) > (0, 0), resolution

    filters: list[str] = []

    # scale (the slowest)
    if resolution is not None and (src := get_resolution(video)) != resolution:
        filters = [f"scale=h={resolution[0]}:w={resolution[1]}:sws_flags=bicubic"]

    # format (medium)
    if pix_fmt is not None and (src := get_pix_fmt(video)) != pix_fmt:
        match = re.search(r"yuv(?P<samp>\d{3})p(?P<bit>\d+)le", src + "8le")
        bit_per_bloc_src = float(match["bit"]) * sum(map(int, match["samp"]))
        match = re.search(r"yuv(?P<samp>\d{3})p(?P<bit>\d+)le", pix_fmt + "8le")
        bit_per_bloc_dst = float(match["bit"]) * sum(map(int, match["samp"]))
        if bit_per_bloc_dst < bit_per_bloc_src:
            filters.insert(0, f"format={pix_fmt}")
        else:
            filters.append(f"format={pix_fmt}")

    # fps (the fastest)
    if fps is not None and (src := get_rate_video(video)) != fps:
        if fps < src:
            filters.insert(0, f"fps={fps}")
        else:
            filters.append(f"fps={fps}")

    # additional filter
    if additional_filter:
        filters.insert(0, additional_filter)

    return ",".join(filters)



"""Use psutil to record the CPU and RAM usage."""

import numbers
import threading
import time

import numpy as np
import psutil


class Usage(threading.Thread):
    """Use psutil through a python context manager.

    Examples
    --------
    >>> import time
    >>> from mendevi.measures.psutil import Usage
    >>> with Usage() as usage:
    ...     time.sleep(1)
    ...
    >>>

    """

    def __init__(self, sleep: numbers.Real = 50e-3) -> None:
        """Initialize the usage context.

        Parameters
        ----------
        sleep : float, default=50e-3
            The time interval between 2 measures (in s).

        """
        super().__init__(daemon=True)

        assert isinstance(sleep, numbers.Real), sleep.__class__.__name__
        assert sleep > 0, sleep

        self._stop_flag = False
        self.sleep = float(sleep)
        self.res: dict = {"dt": [], "pcu": None, "cpus": [], "ram": [], "temp": {}}

    def run(self) -> None:
        """Perform the measures."""
        while not self._stop_flag:
            t_init = time.time()
            self.res["cpus"].append(psutil.cpu_percent(self.sleep, percpu=True))
            self.res["ram"].append(psutil.virtual_memory().used)
            temp: dict[str, float] = {
                f"{generic}_{shwtemp.label or i}": shwtemp.current
                for generic, shwtemps in psutil.sensors_temperatures().items()
                for i, shwtemp in enumerate(shwtemps)
            }
            for lbl, degrees in temp.items():
                self.res["temp"][lbl] = self.res["temp"].get(lbl, [])
                self.res["temp"][lbl].append(degrees)
            self.res["dt"].append(time.time() - t_init)

    def __enter__(self) -> dict:
        """Start to measure.

        Returns
        -------
        Consumption: dict[str]
            * 'cpu': The mean cummulated usage of all the logical cpus.
            * 'cpus': Each logical cpu usage at each time (in %).
            * 'dt': The time difference between 2 consecutive measurements (in s).
            * 'ram': The virtual ram usage (in bytes).
            * 'temp': The temperatures measured at each moment for all sensors (in C).

        """
        self.start()
        return self.res

    def __exit__(self, *_: object) -> None:
        """Stop the measure and update the dictionary returnd by __enter__."""
        self._stop_flag = True
        self.join()  # wait the last update of self.run
        self.res["cpu"] = float(np.mean(self.res["cpus"], axis=0).sum())



"""Interacted with transmission-daemon."""

import logging
import pathlib
import subprocess
import time

import orjson
import tqdm
from context_verbose import Printer


def add_torrent(
    torrent: pathlib.Path, download_dir: pathlib.Path | None = None,
) -> pathlib.Path:
    """Add the torrent to transmission.

    Parameters
    ----------
    torrent : pathlib.Path
        The torrent file to be added.
    download_dir : pathlib.Path, default=~/.cache/mendevi/
        The download_dir in which to download the torrent content.

    Returns
    -------
    file : pathlib.Path
        Return the downloaded file path, related to the torrent.

    """
    # verifications
    assert isinstance(torrent, pathlib.Path), torrent.__class__.__name__
    assert torrent.is_file(), torrent
    if download_dir is None:
        download_dir = pathlib.Path.home() / ".cache" / "mendevi"
        download_dir.mkdir(mode=0o777, parents=True, exist_ok=True)
    else:
        assert isinstance(download_dir, pathlib.Path), download_dir.__class__.__name__
        assert download_dir.is_dir(download_dir)

    # shortcut
    final_file = download_dir / torrent.stem
    if (final_file).exists():
        return final_file

    # add torrent
    test_transmission_is_installed()  # after shortcut
    with Printer(f"Add {torrent.name!r} to transmission...", color="green") as prt:
        try:
            out = subprocess.run(
                [
                    "transmission-remote",
                    "--add", str(torrent),
                    "--download-dir", str(download_dir),
                ],
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as err:
            msg = "the transmission-daemon package appears to be corrupted"
            raise ImportError(msg) from err
        prt.print(out.stdout.decode().strip())

    # download content
    wait_until_finished(torrent.stem)

    return final_file


def test_transmission_is_installed() -> None:
    """Ensures that transmission-daemon has been installed correctly.

    Raises
    ------
    ImportError
        If transmission-daemon is not well installed

    """
    try:
        out = subprocess.run(["transmission-remote", "--version"], check=True, capture_output=True)
    except FileNotFoundError as err:
        msg = (
            "please install transmission-daemon "
            "(https://mendevi.readthedocs.io/latest/developer_guide/installation.html#transmission)"
        )
        raise ImportError(
            msg,
        ) from err
    if not (out.stdout or out.stderr):
        msg = "the transmission-daemon package appears to be corrupted"
        raise logging.getLogger(__name__).warning(msg)


def wait_until_finished(name: str) -> None:
    """Display the progress bar of the torrent download."""
    with Printer(f"Download {name!r} (transmission-remote -l)...", color="green") as prt:
        load = tqdm.tqdm(
            dynamic_ncols=True,
            leave=True,
            smoothing=1e-6,
            total=None,
            unit="Mo",
        )
        while True:
            out = subprocess.run(
                ["transmission-remote", "--json", "--list"],
                check=True,
                capture_output=True,
            )
            try:
                states = orjson.loads(out.stdout)
            except orjson.JSONDecodeError as err:
                msg = f"the transmission-daemon package appears to be corrupted, {out.stdout}"
                raise ImportError(
                    msg,
                ) from err
            if len(state := [t for t in states["arguments"]["torrents"] if t["name"] == name]) == 0:
                msg = f"the torrent {name} is not in the transmission-remote --list"
                raise KeyError(msg)
            state = state.pop()
            if state["errorString"]:
                load.set_description(state["errorString"])
            else:
                load.set_description(f"downloads from {state['peersSendingToUs']} pairs")
            load.total = state["sizeWhenDone"]//1_000_000
            load.update(load.total - state["leftUntilDone"]//1_000_000 - load.n)
            if state["leftUntilDone"] == 0:
                load.close()
                break
            time.sleep(1.0)
        prt.print_time()



"""Help to get the good extractor."""

import ast
import collections
import enum
import importlib
import pathlib
import tempfile
import typing
import uuid

import networkx as nx

from mendevi.database import extract


class Scale(enum.Flag):
    """Represent the nature of the variable, continuous or discrete."""

    LINEAR = enum.auto()
    LOGARITHMIC = enum.auto()
    DISCRETE = enum.auto()


class Extractor(collections.abc.Mapping):
    """Correctly define a variable in a context.

    Examples
    --------
    >>> from mendevi.database.meta import Scale, Extractor
    >>> Extractor("legend", "func", Scale.DISCRETE)
    Extractor('legend', 'func', Scale.DISCRETE)
    >>> {**_}
    {'legend': 'legend', 'func': 'func', 'scale': <Scale.DISCRETE: 4>}
    >>>

    """

    def __init__(self, legend: str, func: typing.Callable | str, scale: Scale) -> None:
        """Create a dictlike named tuple."""
        assert isinstance(legend, str), legend.__class__.__name__
        assert callable(func) or isinstance(func, str), func.__class__.__name__
        assert isinstance(scale, Scale), scale.__class__.__name__
        self._legend, self._func, self._scale = legend, func, scale

    @property
    def legend(self) -> str:
        """Return the description."""
        return self._legend

    @property
    def func(self) -> typing.Callable | str:
        """Return the function or the expression to compute the variable."""
        return self._func

    @property
    def scale(self) -> Scale:
        """Return the kind of variable."""
        return self._scale

    def __getitem__(self, item: str) -> object:
        """Return the attribute."""
        return {
            "legend": self._legend,
            "func": self._func,
            "scale": self._scale,
        }[item]

    def __iter__(self) -> tuple:
        """Yield the keys."""
        yield from ("legend", "func", "scale")

    def __len__(self) -> int:
        """Return the numbers of attributes."""
        return 3

    def __repr__(self) -> str:
        """Return a nice representation."""
        return f"{self.__class__.__name__}({self._legend!r}, {self._func!r}, {self._scale})"


ALL_EXTRACTORS = {
    "act_duration": Extractor(
        "Video processing activity duration in seconds",
        extract.extract_act_duration,
        Scale.LINEAR,
    ),
    "bitrate": Extractor(
        r"Video bitrate in $bit.s^{-1}$",
        "None if size is None or video_duration is None else 8.0 * size / video_duration",
        Scale.LOGARITHMIC,
    ),
    "category": Extractor(
        "3GPP TR 26.955, 5G Video Codec Characteristics",
        (
            "{\n"
            '    "aov": "Gaming",\n'  # ; Key: S5-R01",\n'
            '    "baolei_balloon": "Gaming",\n'  # ; Key: S5-R04",\n'
            '    "baolei_man": "Gaming",\n'  # ; Key: S5-R02",\n'
            '    "baolei_woman": "Gaming",\n'  # ; Key: S5-R03",\n'
            '    "baolei_yard": "Gaming",\n'  # ; Key: S5-R05",\n'
            '    "boat": "4K-TV",\n'  # ; Key: S5-R5",\n'
            '    "bode_museum": "8K-TV",\n'  # ; Key: S3-R16",\n'
            '    "brest_sedof": "4K-TV",\n'  # ; Key: S5-R5",\n'
            '    "cosmos": "4K-TV",\n'  # ; Key: S5-R5",\n'
            '    "cs_go": "Gaming",\n'  # ; Key: S5-R12",\n'
            '    "elevator": "4K-TV",\n'  # ; Key: S5-R5",\n'
            '    "fountain": "4K-TV",\n'  # ; Key: S5-R5",\n'
            '    "graphics_mix_simple": "ScreenContent",\n'  # ; Key: S3-R09",\n'
            '    "graphics_mix_transitions": "ScreenContent",\n'  # ; Key: S3-R13",\n'
            '    "heroes_of_the_storm": "Gaming",\n'  # ; Key: S5-R08",\n'
            '    "jianling_beach": "Gaming",\n'  # ; Key: S5-R07",\n'
            '    "jianling_temple": "Gaming",\n'  # ; Key: S5-R06",\n'
            '    "life_untouched": "4K-TV",\n'  # ; Key: S5-R5",\n'
            '    "meridian": "4K-TV",\n'  # ; Key: S5-R5",\n'
            '    "mine_craft": "Gaming",\n'  # ; Key: S5-R11",\n'
            '    "mission_control": "ScreenContent",\n'  # ; Key: S3-R17",\n'
            '    "mooving_text": "ScreenContent",\n'  # ; Key: S3-R01",\n'
            '    "neon": "Messaging",\n'  # ; Key: S3-R16",\n'
            '    "neptune_fountain_2": "8K-TV",\n'  # ; Key: S3-R16",\n'
            '    "neptune_fountain_3": "8K-TV",\n'  # ; Key: S3-R16",\n'
            '    "nocturne": "4K-TV",\n'  # ; Key: S5-R5",\n'
            '    "oberbaum_spree": "8K-TV",\n'  # ; Key: S3-R16",\n'
            '    "park_joy": "4K-TV",\n'  # ; Key: S5-R5",\n'
            '    "project_cars": "Gaming",\n'  # ; Key: S5-R09",\n'
            '    "quadriga_tree": "8K-TV",\n'  # ; Key: S3-R16",\n'
            '    "rain_fruits": "4K-TV",\n'  # ; Key: S5-R5",\n'
            '    "riverbank": "4K-TV",\n'  # ; Key: S5-R5",\n'
            '    "skater": "Messaging",\n'  # ; Key: S3-R16",\n'
            '    "soccer": "4K-TV",\n'  # ; Key: S5-R5",\n'
            '    "sol_levante": "4K-TV",\n'  # ; Key: S5-R5",\n'
            '    "sparks": "4K-TV",\n'  # ; Key: S5-R5",\n'
            '    "star_craft": "Gaming",\n'  # ; Key: S5-R13",\n'
            '    "subway_tree": "8K-TV",\n'  # ; Key: S3-R16",\n'
            '    "text_mix_transition": "ScreenContent",\n'  # ; Key: S3-R05",\n'
            '    "tiergarten_parkway": "8K-TV",\n'  # ; Key: S3-R16",\n'
            '    "tunnel_flag": "4K-TV",\n'  # ; Key: S5-R5",\n'
            '    "world_of_warcraft": "Gaming",\n'  # ; Key: S5-R10",\n'
            "}.get(reference_video_name, reference_video_name)"
        ),
        Scale.DISCRETE,
    ),
    "codec": Extractor(
        "Codec name",
        extract.extract_codec,
        Scale.DISCRETE,
    ),
    "cores": Extractor(
        "Average cumulative utilisation rate of logical cores",
        extract.extract_cores,
        Scale.LINEAR,
    ),
    "decode_cmd": Extractor(
        "The ffmpeg command used for decoding",
        extract.extract_decode_cmd,
        Scale.DISCRETE,
    ),
    "decode_scenario": Extractor(
        "Unique string specific to the decoding scenario",
        'f"cmd: {decode_cmd}, hostname: {hostname}"',
        Scale.DISCRETE,
    ),
    "effort": Extractor(
        "Effort provided as a parameter to the encoder",
        extract.extract_effort,
        Scale.DISCRETE,
    ),
    "encode_cmd": Extractor(
        "The ffmpeg command used for encoding",
        extract.extract_encode_cmd,
        Scale.DISCRETE,
    ),
    "encode_scenario": Extractor(
        "Unique string specific to the encoding scenario",
        'f"cmd: {encode_cmd}, video_name: {video_name}, hostname: {hostname}"',
        Scale.DISCRETE,
    ),
    "encoder": Extractor(
        "Name of the encoder",
        extract.extract_encoder,
        Scale.DISCRETE,
    ),
    "energy": Extractor(
        "Total energy consumption in Joules",
        "None if powers is None else float((powers[0] * powers[1]).sum())",
        Scale.LOGARITHMIC,
    ),
    "energy_per_frame": Extractor(
        "Average energy consumption per frame in Joules",
        "energy / nbr_frames",
        Scale.LOGARITHMIC,
    ),
    "frames": Extractor(
        "The metadata of each frame",
        extract.extract_frames,
        Scale.DISCRETE,
    ),
    "gamut": Extractor(
        "The tristimulus primaries colors name",
        extract.extract_gamut,
        Scale.DISCRETE,
    ),
    "height": Extractor(
        "Height of images in pixels",
        extract.extract_height,
        Scale.LINEAR,
    ),
    "hostname": Extractor(
        "The machine name",
        extract.extract_hostname,
        Scale.DISCRETE,
    ),
    "lpips": Extractor(
        "Learned Perceptual Image Patch Similarity (LPIPS)",
        extract.extract_lpips,
        Scale.LINEAR,
    ),
    "lpips_alex": Extractor(
        "Learned Perceptual Image Patch Similarity (LPIPS) with alex",
        extract.extract_lpips_alex,
        Scale.LINEAR,
    ),
    "lpips_vgg": Extractor(
        "Learned Perceptual Image Patch Similarity (LPIPS) with vgg",
        extract.extract_lpips_vgg,
        Scale.LINEAR,
    ),
    "power": Extractor(
        "Average power consumption in Watts",
        "None if energy is None or powers is None else energy / float(powers[0].sum())",
        Scale.LINEAR,
    ),
    "powers": Extractor(
        "The interval duration and the average power in each intervals",
        extract.extract_powers,
        Scale.DISCRETE,
    ),
    "mode": Extractor(
        "Bitrate mode, constant (cbr) or variable (vbr)",
        extract.extract_mode,
        Scale.DISCRETE,
    ),
    "nbr_frames": Extractor(
        "The real number of frames of the video file",
        "None if frames is None else len(frames)",
        Scale.LOGARITHMIC,
    ),
    "profile": Extractor(
        "Profile of the video",
        (
            "None if height is None and width is None else "
            "best_profile(height or width, width or height)"
        ),
        Scale.DISCRETE,
    ),
    "psnr": Extractor(
        "Peak Signal to Noise Ratio (PSNR)",
        extract.extract_psnr,
        Scale.LINEAR,
    ),
    "quality": Extractor(
        "Quality level passed to the encoder",
        extract.extract_quality,
        Scale.LINEAR,
    ),
    "range": Extractor(
        "Video encoding color range, 'tv' or 'pc'",
        extract.extract_range,
        Scale.DISCRETE,
    ),
    "reference_video_name": Extractor(
        "Input video basename",
        extract.extract_reference_video_name,
        Scale.DISCRETE,
    ),
    "rms_sobel": Extractor(
        "Spacial root mean square sobel gradient complexity",
        extract.extract_rms_sobel,
        Scale.LINEAR,
    ),
    "rms_time_diff": Extractor(
        "Temporal root means square time difference complexity",
        extract.extract_rms_time_diff,
        Scale.LINEAR,
    ),
    "shape": Extractor(
        "The image shapes height x width in pixels",
        "(height, width)",
        Scale.DISCRETE,
    ),
    "ssim": Extractor(
        "Structural Similarity (SSIM)",
        extract.extract_ssim,
        Scale.LINEAR,
    ),
    "ssim_comp": Extractor(
        "Complementary of Structural Similarity (1-SSIM)",
        "None if ssim is None else 1.0 - ssim",
        Scale.LOGARITHMIC,
    ),
    "temp": Extractor(
        "Average temperature in C",
        extract.extract_temp,
        Scale.LINEAR,
    ),
    "threads": Extractor(
        "Number of threads provided as a parameter to the encoder",
        extract.extract_threads,
        Scale.LINEAR,
    ),
    "transfer": Extractor(
        "The non-linear transfer function name",
        extract.extract_transfer,
        Scale.DISCRETE,
    ),
    "vmaf": Extractor(
        "Video Multi-Method Assessment Fusion (VMAF)",
        extract.extract_vmaf,
        Scale.LINEAR,
    ),
    "video_duration": Extractor(
        "Video duration in seconds",
        extract.extract_video_duration,
        Scale.LINEAR,
    ),
    "video_hash": Extractor(
        "The hexadecimal md5 video file checksum",
        extract.extract_video_hash,
        Scale.DISCRETE,
    ),
    "video_name": Extractor(
        "Video basename",
        extract.extract_video_name,
        Scale.DISCRETE,
    ),
    "video_size": Extractor(
        "The total video file size in bytes",
        extract.extract_video_size,
        Scale.LOGARITHMIC,
    ),
    "width": Extractor(
        "Width of images in pixels",
        extract.extract_width,  # comme dans ton code
        Scale.LINEAR,
    ),
}


ALIAS = {
    "rate": "bitrate",
    "dec_cmd": "decode_cmd",
    "actdur": "act_duration",
    "prim": "gamut",
    "primaries": "gamut",
    "color_primaries": "gamut",
    "dec_scenario": "decode_scenario",
    "preset": "effort",
    "enc_cmd": "encode_cmd",
    "enc_scenario": "encode_scenario",
    "vid_duration": "video_duration",
    "video_md5": "video_hash",
    "vid_md5": "video_hash",
    "vid_hash": "video_hash",
    "vid_name": "video_name",
    "name": "video_name",
    "vid_size": "video_size",
    "size": "video_size",
    "ref_name": "reference_video_name",
    "reference_vid_name": "reference_video_name",
    "reference_name": "reference_video_name",
    "ref_vid_name": "reference_video_name",
    "ref_video_name": "reference_video_name",
    "nb_frames": "nbr_frames",
    "psnr_rev": "ssim_comp",
    "comp_ssim": "ssim_comp",
    "ssim_rev": "ssim_comp",
    "rev_ssim": "ssim_comp",
    "temperature": "temp",
    "trans": "transfer",
    "color_transfer": "transfer",
    "eotf": "transfer",
}


def _import_extractor(code: list[str]) -> callable:
    """Import the function line_extractor."""
    code = [
        "from mendevi.utils import best_profile",
        "import mendevi.database.extract as extract",
        "",
        *code,
    ]
    path = pathlib.Path(tempfile.gettempdir()) / f"{uuid.uuid4().hex}.py"
    with path.open("w", encoding="utf-8") as file:
        file.write("\n".join(code))
    spec = importlib.util.spec_from_file_location(path.stem, path)
    modulevar = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(modulevar)
    # path.unlink()
    return modulevar.line_extractor


def _print_branch(leaf: str, skip: set[str], graph: nx.DiGraph) -> list[str]:
    """Return the code to extract a branch."""
    branch_code: list[str] = []
    predecessors = {leaf} - skip
    while predecessors:
        skip |= predecessors
        for lbl in sorted(predecessors):  # sorted for reproductibility and lisibility
            legend, func = graph.nodes[lbl]["legend"], graph.nodes[lbl]["func"]
            branch_code.insert(
                0,
                (
                    f"    {lbl} = extract.{func.__name__}(raw)  # {legend}"
                    if callable(func) else
                    f"    {lbl} = {func}  # {legend}"
                ),
            )
        predecessors = {pp for p in predecessors for pp in graph.predecessors(p)} - skip
    return branch_code


def extract_names(expr: str) -> set[str]:
    """Return all the symbols in the python expression.

    Examples
    --------
    >>> from mendevi.database.meta import extract_names
    >>> extract_names("foo")
    {'foo'}
    >>> extract_names("[i**2 for i in foo]")
    {'foo'}
    >>> extract_names("foo.bar")
    {'foo'}
    >>> extract_names("bar(foo)")
    {'foo'}
    >>> extract_names("foo.bar()")
    {'foo'}
    >>>

    """
    try:
        nodes = list(ast.walk(ast.parse(expr, mode="exec")))
    except SyntaxError as err:
        msg = f"the argument {expr!r} is not a valid python expression"
        raise SyntaxError(
            msg,
        ) from err
    reject = {
        n.id for n in nodes if isinstance(n, ast.Name) and isinstance(n.ctx, ast.Store | ast.Del)
    } | {
        n_.id
        for n in nodes if isinstance(n, ast.Call) and not isinstance(n.func, ast.Attribute)
        for n_ in ast.walk(n.func) if isinstance(n_, ast.Name)
    }
    candidates = {n.id for n in nodes if isinstance(n, ast.Name)}
    return set(candidates - reject)  # set usefull for empty case


def get_extractor(name: str, *, safe: bool = False) -> Extractor:
    """Get the way to deserialize a raw value.

    Parameters
    ----------
    name : str
        The label name.
    safe : boolean, default=False
        If True, retrun a stupid value instead of raising KeyError.

    Returns
    -------
    extractor : Extractor
        The label, func and scale.

    """
    assert isinstance(name, str), name.__class__.__name__
    assert isinstance(safe, bool), safe.__class__.__name__
    if (extractor := ALL_EXTRACTORS.get(ALIAS.get(name, name), None)) is not None:
        return extractor
    if safe:
        return Extractor(name, name, Scale.LINEAR)
    msg = f"{name} is not recognised"
    raise KeyError(msg)


def merge_extractors(
    labels: set[str],
    alias: dict[str, str] | None = None,
    select: str | None = None,
    *,
    return_callable: bool = False,
) -> tuple[set[str], list[str] | typing.Callable]:
    r'''Return the source code of the function that extracts all variables.

    Parameters
    ----------
    labels : set[str]
        The returned variable names. These are the keys to the output dictionary.
    alias : dict[str, str], optional
        To each new name, associate an expression.
        By default, the label extraction method is defined by the function :py:func:`get_extractor`.
        This list of aliases allows any unknown key to define a customised access method.
    select : str, optional
        A Python Boolean expression that raises a RejectError exception if it evaluates to False.
    return_callable : boolean, default=False
        By default, returns the source code of the function.
        If this option is set to True, an executable function is returned.

    Returns
    -------
    lbls_atom : set[str]
        The name of the primary value to be extracted for the SQL query.
    func : list[str] or callable
        The function that consumes a line from the SQL query,
        and returns the dictionary of extracted values.

    Examples
    --------
    >>> from mendevi.database.meta import merge_extractors
    >>> labels = {"enc_scenario", "rate", "power", "foo"}
    >>> alias = {"foo": "(abaca, energy)", "abaca": "(hostname, nbr_frames, enc_scenario)"}
    >>> select = '"yeti" not in hostname'
    >>> print("\n".join(merge_extractors(labels, alias, select)[1]))  # doctest: +ELLIPSIS
    def line_extractor(raw: dict[str]) -> dict[str]:
        """Get the labels: enc_scenario, foo, power, rate.
    <BLANKLINE>
        Causality constraint:
            * abaca -> foo
            * enc_scenario -> abaca
            * encode_cmd -> enc_scenario
            * energy -> foo
            * energy -> power
            * frames -> nbr_frames
            * hostname -> abaca
            * hostname -> enc_scenario
            * hostname -> reject
            * nbr_frames -> abaca
            * powers -> energy
            * powers -> power
            * size -> rate
            * video_duration -> rate
            * video_name -> enc_scenario
        """
        # reject wrong line
        hostname = extract.extract_hostname(raw)  # The machine name
        reject = not ("yeti" not in hostname)  # not ("yeti" not in hostname)
        if reject:
            msg = "this line must be filtered"
            raise RejectError(msg)
    <BLANKLINE>
        # extract data
        video_name = extract.extract_video_name(raw)  # Video basename
        encode_cmd = extract.extract_encode_cmd(raw)  # The ffmpeg command used for encoding
        enc_scenario = f"cmd: {encode_cmd}, video_name: {video_name}, hostname: {hostname}"  # Un...
        frames = extract.extract_frames(raw)  # The metadata of each frame
        powers = extract.extract_powers(raw)  # The interval duration and the average power in ea...
        nbr_frames = None if frames is None else len(frames)  # The real number of frames of the ...
        energy = None if powers is None else float((powers[0] * powers[1]).sum())  # Total energy...
        abaca = (hostname, nbr_frames, enc_scenario)  # alias
        foo = (abaca, energy)  # alias
        power = energy / float(powers[0].sum())  # Average power consumption in Watts
        video_duration = extract.extract_video_duration(raw)  # Video duration in seconds
        size = extract.extract_video_size(raw)  # The total video file size in bytes
        rate = None if size is None or video_duration is None else 8.0 * size / video_duration  #...
    <BLANKLINE>
        return {
            'enc_scenario': enc_scenario,
            'foo': foo,
            'power': power,
            'rate': rate,
        }
    >>>

    '''
    assert isinstance(labels, set), labels.__class__.__name__
    assert all(isinstance(lbl, str) for lbl in labels), labels
    alias = alias or {}
    assert isinstance(alias, dict), alias.__class__.__name__
    assert select is None or isinstance(select, str), select.__class__.__name__

    extractors = nx.DiGraph()

    def get_alias_extractor(alias: dict, lbl: str) -> Extractor:
        return (
            Extractor("alias", alias[lbl], get_extractor(alias[lbl], safe=True).scale)
            if lbl in alias else
            get_extractor(lbl, safe=True)
        )

    # initialise the graph with final leaves
    for label in labels:
        extractors.add_node(label, **get_alias_extractor(alias, label))
    # extractors.add_nodes_from(labels)  # adding leaves
    if select is not None:
        extractors.add_node(
            "reject", **get_extractor(f"not ({select})", safe=True),
        )

    # construct the full tree
    while nodes := [
        n for n, deg in extractors.in_degree()
        if deg == 0 and not callable(extractors.nodes[n]["func"])
    ]:
        for node in nodes:
            for root in extract_names(alias.get(node, extractors.nodes[node]["func"])):
                extractors.add_node(root, **get_alias_extractor(alias, root))
                extractors.add_edge(root, node)

    # # draw graph for debug: sudo apt install graphviz graphviz-dev && uv pip install pygraphviz
    # with open("/tmp/extractors.dot", "w") as file:
    #     for node in extractors:
    #         extractors.nodes[node]["label"] = (
    #             f"{node}\\n{'\\n'.join(f'{k}:{v!r}' for k, v in extractors.nodes[node].items())}"
    #         )
    #     file.write(nx.nx_agraph.to_agraph(extractors).string())

    # verification of undefinded variable or cycle
    if cycles := list(nx.simple_cycles(extractors)):
        msg = (
            f"The extraction graph has cycles: {cycles}, "
            "which means that variables are not defined. "
            "You must specify an alias to break all cycles."
        )
        raise ValueError(msg)

    # create the source code
    # 1) header
    code = [
        "def line_extractor(raw: dict[str]) -> dict[str]:",
        f'    """Get the labels: {", ".join(sorted(labels))}.',
        "",
        "    Causality constraint:",
        *(f"        * {prev} -> {node}" for prev, node in sorted(extractors.edges)),
        '    """',
    ]
    skip: set[str] = set()  # processed nodes

    # 2) reject bad line
    if select is not None:
        code.extend([
            "    # reject wrong line",
            *_print_branch("reject", skip, extractors),
            "    if reject:",
            '        msg = "this line must be filtered"',
            "        raise RejectError(msg)",
            "",
        ])

    # 3) extract the tree for all labels
    code.extend([
        "    # extract data",
        *(line for leaf in sorted(labels) for line in _print_branch(leaf, skip, extractors)),
        "",
    ])

    # 4) cast and return the values
    code.extend([
        "    return {",
        *(f"        {lbl!r}: {lbl}," for lbl in sorted(labels)),
        "    }",
    ])

    # import source code
    return (
        {n for n in extractors if callable(extractors.nodes[n]["func"])},
        (_import_extractor(code) if return_callable else code),
    )



"""Allow to handle ffmpeg command line."""

import pathlib
import shlex


class CmdFFMPEG:
    """Allow easy manipulation of a complete ffmpeg expression.

    ffmpeg -y -hide_banner -loglevel verbose

    Attributes
    ----------
    decode : list[str]
        The options between -c:v and -i (read and write).
    vid_filter : str
        The filter string after -vf (read and write).
    general : list[str]
        The options immediately after ffmpeg and immediately before the decoder (read and write).
    video : pathlib.Path
        The input video path (readonly).
    encode : list[str]
        The encoder name and options after -c:v (read and write).
    output : list[str]
        The final arguments of the ffmpeg cmd (read and write).

    """

    def __init__(
        self,
        video: pathlib.Path | str,
        **kwargs: dict[str],
    ) -> None:
        """Initialise the ffmpeg cmd.

        Parameters
        ----------
        video : pathlike
            The input video path.
        **kwargs: dict
            See above.
        general : list[str], str, optional
            The options immediately after ffmpeg and immediately before the decoder.
        decode : list[str], str, optional
            The options between ``-c:v`` and ``-i``.
        vid_filter : str, optional
            The filter string after ``-vf``.
        encode: list[str], str, optional
            The encoder name and options after ``-c:v``.
        output: list[str], str, default="-f null -"
            The final arguments of the ffmpeg command, after the encoder description.

        """
        # initialisation
        video = pathlib.Path(video).expanduser()
        self._video = video
        self._general = self._decode = self._vid_filter = self._encode = self._output = None

        # parse and check, using setter
        self.general = kwargs.get("general")
        self.decode = kwargs.get("decode")
        self.vid_filter = kwargs.get("vid_filter", "")
        self.encode = kwargs.get("encode")
        self.output = kwargs.get("output")

    @property
    def decode(self) -> list[str]:
        """Return the options between -c:v and -i."""
        return self._decode.copy()

    @decode.setter
    def decode(self, decode: list[str] | str | None) -> None:
        """Update the options between -c:v and -i."""
        match decode:
            case None:
                self._decode = []
            case str():
                self._decode = shlex.split(decode)
            case list():
                assert all(isinstance(cmd, str) for cmd in decode), decode
                self._decode = decode.copy()
            case _:
                msg = f"'decode' has to be None, str or list, not {decode.__class__.__name__}"
                raise TypeError(msg)

    @property
    def encode(self) -> list[str]:
        """Return the encoder name and options after -c:v."""
        return self._encode.copy()

    @encode.setter
    def encode(self, encode: list[str] | str | None) -> None:
        """Update the encoder name and options after -c:v."""
        match encode:
            case None:
                self._encode = []
            case str():
                self._encode = shlex.split(encode)
            case list():
                assert all(isinstance(cmd, str) for cmd in encode), encode
                self._encode = encode.copy()
            case _:
                msg = f"'encode' has to be None, str or list, not {encode.__class__.__name__}"
                raise TypeError(msg)

    @property
    def output(self) -> list[str]:
        """Return the final arguments of the ffmpeg command, after the encoder."""
        return self._output.copy()

    @output.setter
    def output(self, output: list[str] | str | None) -> None:
        """Update the final arguments of the ffmpeg command, after the encoder."""
        match output:
            case None:
                self._output = ["-f", "null", "-"]
            case str():
                self._output = shlex.split(output)
            case list():
                assert all(isinstance(cmd, str) for cmd in output), output
                self._output = output.copy()
            case _:
                msg = f"'output' has to be None, str or list, not {output.__class__.__name__}"
                raise TypeError(msg)

    @property
    def vid_filter(self) -> str:
        """Return the filter string after -vf."""
        return self._filter

    @vid_filter.setter
    def vid_filter(self, vid_filter: str) -> None:
        """Update the filter string after -vf."""
        assert isinstance(vid_filter, str), vid_filter.__class__.__name__
        self._vid_filter = vid_filter

    @property
    def general(self) -> list[str]:
        """Return the options immediately after ffmpeg and immediately before the decoder."""
        if self._general is None:
            return ["-y", "-loglevel", "verbose"]
        return self._general.copy()

    @general.setter
    def general(self, general: list[str] | str | None) -> None:
        """Update the options immediately after ffmpeg and immediately before the decoder."""
        match general:
            case None:
                self._general = None
            case str():
                self._general = shlex.split(general)
            case list():
                assert all(isinstance(cmd, str) for cmd in general), general
                self._general = general.copy()
            case _:
                msg = f"'general' has to be None, str or list, not {general.__class__.__name__}"
                raise TypeError(msg)

    @property
    def video(self) -> pathlib.Path:
        """Return the input video path."""
        return self._video

    def __iter__(self) -> str:
        """Iterate over each parameter to be compatible with list(self)."""
        yield "ffmpeg"
        yield from self.general
        if self._decode:
            yield "-c:v"
            yield from self._decode
        yield "-i"
        yield str(self._video)
        if self._vid_filter:
            yield "-vf"
            yield self._vid_filter
        if self._encode:
            yield "-c:v"
            yield from self._encode
        yield from self._output

    def __str__(self) -> str:
        """Return the full shell cmd.

        Examples
        --------
        >>> from mendevi.cmd import CmdFFMPEG
        >>> print(CmdFFMPEG(video="src.mp4"))
        ffmpeg -y -loglevel verbose -i src.mp4 -f null -
        >>>

        """
        return " ".join(map(shlex.quote, self))



"""CLI entry point to visualize database data."""

import importlib
import pathlib

import click
from context_verbose import Printer
from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers import PythonLexer

from mendevi.cst.labels import LABELS
from mendevi.database.meta import get_extractor
from mendevi.download.decapsulation import retrive_file
from mendevi.plot.printer import printer, safe_lbl

from .parse import parse_expr, parse_videos_database

NAMES_DOC = "        ".join(
    f"* {n}: {get_extractor(n).legend}.\n" for n in LABELS
)
DOCSTRING = f"""Draw a chart from a database.

\b
Parameters
----------
database : pathlike, optional
    The path to the database where all measurements are stored.
    If a folder is provided, the database is created inside this folder.
x, y : tuple[str]
    The name of the x and y axes on each window. Possible values are:
        {NAMES_DOC}
\b
color, marker : str, optional
    All points with the same value for this variable will be displayed in the same color and marker.
    Conversely, this allows points that do not share this criterion to be visually separated.
    The possible values are the same as for the x and y parameters.
error : str, optional
    Allows you to average several points within the same error bar.
    The string provided is a Python expression or variable.
    All points with the same image across this function are grouped together.
window : str, optional
    Rather than putting everything in the same chart, this allows you to create sub-windows.
    There will be as many sub-windows as there are different values taken by this criterion.
    The possible values are the same as for the x and y parameters.
window_x, window_y : str, optional
    Same as window, but with precision on the axis on which to develop.
filter : str, optional
    This allows you to filter the data to select only a portion of it.
    The character string provided must be a Boolean Python expression
    that returns True to keep the data. For example: ``enc_duration > 10 and profile == "fhd"``.
table : str, optional
    The main sql table juste after the FROM in SELECT.
"""


def _parse_args(prt: Printer, kwargs: dict) -> None:
    """Verification of the arguments."""
    # x
    assert "x" in kwargs, sorted(kwargs)
    assert isinstance(kwargs["x"], tuple), kwargs["x"].__class__.__name__
    assert kwargs["x"], "you must provide the -x option"
    for i, expr in enumerate(kwargs["x"]):
        parse_expr(expr, prt, f"x_{i:<8}")

    # y
    assert "y" in kwargs, sorted(kwargs)
    assert isinstance(kwargs["y"], tuple), kwargs["y"].__class__.__name__
    assert kwargs["y"], "you must provide the -y option"
    for i, expr in enumerate(kwargs["y"]):
        parse_expr(expr, prt, f"y_{i:<8}")

    # color
    kwargs["color"] = kwargs.get("color")
    parse_expr(kwargs["color"], prt, "color     ")

    # marker
    kwargs["marker"] = kwargs.get("marker")
    parse_expr(kwargs["marker"], prt, "marker    ")

    # error
    kwargs["error"] = kwargs.get("error")
    parse_expr(kwargs["error"], prt, "error     ")

    # window
    kwargs["window"] = kwargs.get("window")
    if kwargs["window"] is not None:
        kwargs[
            {
                (False, True, False, True): "window_y",
                (True, False, True, False): "window_x",
                (True, True, False, True): "window_y",
                (True, True, True, False): "window_x",
                (False, True, False, False): "window_y",
                (True, False, False, False): "window_x",
            }.get(
                (
                    kwargs.get("window_x") is None,
                    kwargs.get("window_y") is None,
                    len(kwargs["x"]) == 1,
                    len(kwargs["y"]) == 1,
                ),
                "window_y",
            )
        ] = kwargs["window"]

    # window_x
    kwargs["window_x"] = kwargs.get("window_x")
    parse_expr(kwargs["window_x"], prt, "window x  ")

    # window_y
    kwargs["window_y"] = kwargs.get("window_y")
    parse_expr(kwargs["window_y"], prt, "window y  ")

    # filter
    kwargs["filter"] = kwargs.get("filter")
    parse_expr(kwargs["filter"], prt, "filter    ")

    # table
    kwargs["table"] = kwargs.get("table")
    if kwargs["table"] is not None:
        assert isinstance(kwargs["table"], str), kwargs["table"].__class__.__name__
        prt.print(f"table     : {kwargs['table']}")


@click.command(help=DOCSTRING)
@click.argument("database", type=click.Path())
@click.option(
    "-x",
    type=str,
    multiple=True,
    help="The code for the value to be displayed on the x-axis.",
)
@click.option(
    "-y",
    type=str,
    multiple=True,
    help="The code for the value to be displayed on the y-axis.",
)
@click.option(
    "-c", "--color",
    type=str,
    help="The discriminating criterion for colour.",
)
@click.option(
    "-m", "--marker",
    type=str,
    help="The discriminating criterion for marker.",
)
@click.option(
    "-w", "--window",
    type=str,
    help="The criterion on which to separate into sub-windows.",
)
@click.option(
    "-wx", "--window-x",
    type=str,
    help="The criterion on which to separate into sub-windows along x axis.",
)
@click.option(
    "-wy", "--window-y",
    type=str,
    help="The criterion on which to separate into sub-windows along y axis.",
)
@click.option(
    "-e", "--error",
    type=str,
    help="Merge criteria for the error bar.",
)
@click.option(
    "--filter", "-f",
    type=str,
    help="Data selection python expression.",
)
@click.option(
    "-p", "--print",
    is_flag=True,
    help="Flag to print the source code.",
)
@click.option(
    "-t", "--table",
    type=str,
    help="Form the SQL query around this table.",
)
def main(database: str, **kwargs: dict) -> None:
    """See docstring at DOCSTRING."""
    # parse args
    with Printer("Parse configuration...") as prt:
        database = retrive_file(database)
        _, database = parse_videos_database(prt, (), database)
        _parse_args(prt, kwargs)

    # get title name
    path = (
        pathlib.Path.cwd() / (
            f"{'_'.join(map(safe_lbl, kwargs['y']))}"
            "_as_a_function_of_"
            f"{'_'.join(map(safe_lbl, kwargs['x']))}"
            ".py"
        )
    )

    # get code content
    code = printer(database=database, **kwargs, path=path)
    if kwargs.get("print", False):
        print("**********************************PYTHON CODE**********************************")
        print(highlight(code, PythonLexer(), TerminalFormatter()))
        print("*******************************************************************************")
    with path.open("w", encoding="utf-8") as file:
        file.write(code)

    # excecute code
    spec = importlib.util.spec_from_file_location(path.stem, path)
    modulevar = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(modulevar)
    modulevar.main(modulevar.read_sql(database))


# fill the documentation
main.__doc__ = DOCSTRING
